"""
    SecTheory.SecEstimator.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
from molass_legacy._MOLASS.SerialSettings import get_setting
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Models.ElutionCurveModels import egh

NUM_SEC_PARAMS = 6      # used in other modules

def guess_initial_secparams(init_xr_params, rgs):
    Npc = get_setting('num_plates_pc')
    k = np.argmax(init_xr_params[:,0])
    tR = init_xr_params[k,1]
    s, t = init_xr_params[k,[2,3]]
    sigma = np.sqrt(s**2 + t**2)
    """
    tR - tI = sqrt(Np)*sigma
    tI = tR - sqrt(Np)*sigma
    tR = t0 + P*(1 - rho)**m
    t0 = tR - P*(1 - rho)**m
    """
    tI = tR - np.sqrt(Npc)*sigma
    rp = get_setting('poresize')
    print("----------------- guess_initial_secparams: rp=", rp)
    rg = rgs[k]
    rho = rg/rp
    P = 2000
    m = 2
    t0 = tR - P*(1 - rho)**m
    return Npc, rp, tI, t0, P, m


NUM_PARAMS = 4

class SecEstimator:
    def __init__(self, Npc, rp, tI, t0, P, m):
        m = min(2, m)   # force
        self.init_sec_params = Npc, rp, tI, t0, P, m
        print("init: Npc, rp, tI, t0, P, m =", self.init_sec_params)

    def fit_to_decomposition(self, x, y, init_xr_params, rgs, debug=False):
        """
            tR = t0 + P*(1 - rho)**m
            Tr = tR - tI
        """

        n = len(rgs)
        xr_params_shape = (n,4)
        zeros = np.zeros(n)
        Npc, rp = self.init_sec_params[0:2]
        rN = np.sqrt(Npc)
        rhos = rgs/rp
        rhos[rhos > 1] = 1

        def objective(p, debug=False, sub_title=""):
            tI, t0, P, m = p[:NUM_PARAMS]
            xr_params = p[NUM_PARAMS:].reshape(xr_params_shape)

            tR = t0 + P*(1 - rhos)**m
            sigma = (tR - tI)/rN
            s = np.sqrt(xr_params[:,2]**2 + xr_params[:,3]**2)
            scale = np.sum((xr_params[:,0] - init_xr_params[:,0])**2)
            # bound = min(0, tI + 1000)**2 + np.sum(np.min([zeros, sigma], axis=0)**2) + min(0, P - 1000)**2 + max(0, P - 10000)**2
            if debug:
                with plt.Dp():
                    fig, ax = plt.subplots()
                    ax.set_title("objective: " + sub_title)
                    ax.plot(x, y, color="orange")

                    cy_list = []
                    for h_, mu, s_, t_ in xr_params:
                        cy = egh(x, h_, mu, s_, t_)
                        cy_list.append(cy)
                        ax.plot(x, cy, ":")

                    ty = np.sum(cy_list, axis=0)
                    ax.plot(x, ty, ":", color="red")
                    fig.tight_layout()
                    plt.show()

            return scale + np.sum((xr_params[:,1] - tR)**2 + (s - sigma)**2)

        start = 6 - NUM_PARAMS
        init_params = np.concatenate([self.init_sec_params[start:], init_xr_params.flatten()])

        if debug:
            objective(init_params, debug=True, sub_title="before")

        Npc, rp, tI, t0, P, m = self.init_sec_params

        bounds = [
                    (tI - 500, tI + 200),
                    (t0 - 500, t0 + 200),
                    (1000, 10000),
                    (1, 3),
                ]

        for h_, mu, s_, t_ in init_xr_params:
            bounds.append((h_*0.8, h_*1.2))
            ts_ = np.sqrt(s_**2 + t_**2)
            dx = ts_*0.1
            bounds.append((mu - dx, mu + dx))
            bounds.append((s_ - dx, s_ + dx))
            bounds.append((t_ - dx, t_ + dx))
        # ret = minimize(objective, init_params, method='Nelder-Mead', bounds=bounds)
        ret = basinhopping(objective, init_params, minimizer_kwargs=dict(method='Nelder-Mead', bounds=bounds))

        if debug:
            objective(ret.x, debug=True, sub_title="after")

        self.ret_sec_params = np.concatenate([[Npc, rp], ret.x[:NUM_PARAMS]])

        fit_xr_params = ret.x[NUM_PARAMS:].reshape(xr_params_shape)
        # fit_xr_params[:,0] = init_xr_params[:,0]

        if debug:
            print("ret_sec_params=", self.ret_sec_params)
            print("fit_xr_params=", fit_xr_params)

            init_cy_list = []
            for h, tr, s, t in init_xr_params:
                cy = egh(x, h,tr, s, t)
                init_cy_list.append(cy)
            init_ty = np.sum(init_cy_list, axis=0)

            fit_cy_list = []
            for h, tr, s, t in fit_xr_params:
                cy = egh(x, h, tr, s, t)
                fit_cy_list.append(cy)
            fit_ty = np.sum(fit_cy_list, axis=0)

            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("fit_to_decomposition")
                ax.plot(x, y, color="orange")
                for cy in fit_cy_list:
                    ax.plot(x, cy, ":")
                ax.plot(x, init_ty, color="yellow")
                ax.plot(x, fit_ty, ":", color="red")
                fig.tight_layout()
                plt.show()

    def get_params(self):
        return self.ret_sec_params



# coding: utf-8
"""
    InputManager.py

    Copyright (c) 2019-2020, SAXS Team, KEK-PF
"""
from time import sleep
import numpy as np
import threading
from .FolderInfo import FolderInfo

class InputManager:
    def __init__(self, logger=None, debug=False):
        # logger is not None for single process uses
        self.logger = logger
        self.debug = debug

    def manage_loop(self, job_queue, ret_queue, log_folder):
        from ChangeableLogger import Logger
        if self.debug:
            print('manage_loop')
        self.stop = False
        self.interval = 0.5 if self.debug else 0.1
        self.job_queue = job_queue
        self.ret_queue = ret_queue
        self.logger = Logger(log_folder+'/input.log')
        self.logger.info('start')
        self.busy = False
        self.load_thread = None

        while not self.stop:
            if self.debug:
                print('managing')
            if self.job_queue.empty():
                sleep(self.interval)
            else:
                info = self.job_queue.get()
                if info.code == 'stop':
                    break
                self.execute(info)

        self.close()

    def close(self):
        self.logger.info('stop')

    def execute(self, info):
        if info.log:
            self.logger.info('executing %s', info.code)

        try:
            action = self.get_action(info)
            if action is None:
                raise RuntimeError(info)
            else:
                ret = action(info.args)
                if info.wait:
                    self.ret_queue.put(ret)
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            self.logger.error(str(etb))

    def get_action(self, info):
        action = self.__getattribute__(info.code)
        return action

    def start_load_data(self, args):
        print('start_load_data',  args)
        self.busy = True
        self.load_thread = threading.Thread(
                                target=self.load_data,
                                name='LoadDataThread',
                                args=args,
                                )
        self.load_thread.start()
        self.logger.info( 'started loading ' + str(args) )

    def load_data(self, in_folder, for_sec, xray_only, shared=True):
        from SerialDataUtils import load_intensity_files, load_uv_array
        if shared:
            from molass_legacy.KekLib.SharedArrays import SharedArrays

        print('load_data',  in_folder)

        self.sa_list = []

        array, files, comments = load_intensity_files(in_folder, return_comments=True)
        self.logger.info( 'loaded ' + in_folder )
        array_list = [array]
        self.files = files
        self.comments = comments

        if for_sec and not xray_only:
            array, vector, conc_file = load_uv_array(in_folder)
            self.conc_file = conc_file
            array_list += [array, vector]
        else:
            self.conc_file = None

        if shared:
            self.sarrays = SharedArrays(array_list)
        else:
            return array_list

        self.busy = False

    def join_load_thread(self):
        if self.load_thread is not None:
            self.load_thread.join()
            self.load_thread = None

    def is_ready(self, args):
        return not self.busy

    def get_array_info(self, args):
        return self.sarrays.get_name(), self.sarrays.get_tuples()

    def get_folder_info(self, args=None):
        return FolderInfo(self.files, self.conc_file, self.comments)



"""
    Models.Stochastic.ParamLimits.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
USE_K = True

N_BOUND = 300, 3000         # N=372 for 20160227, N>2000 is harmful for OA01
if USE_K:
    KT_BOUND = 500, 2000
else:
    KT_BOUND = 0.01, 3
T0_BOUND = [-1000, +1000]   # these will be changed according to t0_upper in make_mnp_bounds

PORESIZE_INTEG_LIMIT = 600  # changing this value to 600 once seemed harmful in LognormalPoreFunc.py
MAX_PORESIZE = 500      

M1_WEIGHT = 9
M2_WEIGHT = 1

PORESIZE_BOUNDS = 30, MAX_PORESIZE
PORESIZE_MEAN_BOUND = 40, MAX_PORESIZE + 50      # to be changed later
PORESIZE_STDEV_BOUND = 1, 100
LN_MU_BOUND = 3, 7      # exp(3)=20, exp(7)=1097
LN_SIGMA_BOUND = 0.05, 3

MNP_BOUNDS = [N_BOUND, KT_BOUND, T0_BOUND, PORESIZE_BOUNDS]
SCALE_BOUND = 0, 10
RG_BOUND = 5, 100       # upper bound will be later replaced according to reliable_rgmax
TIMESCALE_BOUND = 0.1, 2

BASINHOPPING_SCALE = 10

def make_monopore_bounds(t0_upper, num_scales, num_unreliables, reliable_rgmax, num_timescales):
    MNP_BOUNDS[2] = [t0_upper - 300, t0_upper]
    return np.array(MNP_BOUNDS
                    + [SCALE_BOUND] *num_scales
                    + [(RG_BOUND[0], reliable_rgmax*1.5)]*num_unreliables
                    + [TIMESCALE_BOUND]*num_timescales
                    )

N0_INIT = 14400     # 48000*0.3 (30cm) or (t0/σ0)**2, see meeting document 20221104/index.html 
N0_BOUND = N0_INIT//2, N0_INIT*2
PROP_BOUND = 0, 1

def make_oligopore_bounds(t0_lower, t0_upper, num_pszs, num_scales, num_unreliables, reliable_rgmax, num_timescales, dispersive=False):
    if t0_lower is None:
        t0_lower = t0_upper - 1000
    MNP_BOUNDS[2] = [t0_lower, t0_upper]
    bounds_list = (MNP_BOUNDS[0:3]
                    # + [(20, 50), (50, 100), (100, 500)]
                    + [PORESIZE_BOUNDS]*num_pszs      # allowing the upper bound to reach MAX_PORESIZE can cause problems?
                    + [PROP_BOUND]*(num_pszs-1)
                    )
    bounds_list += [SCALE_BOUND] *num_scales + [(RG_BOUND[0], reliable_rgmax*1.5)]*num_unreliables + [TIMESCALE_BOUND]*num_timescales
    if dispersive:
        bounds_list += [N0_BOUND]
    return np.array(bounds_list)

"""
task: unify V2-scaling using this scaler 
"""
class ParamsScaler:
    def set_scales(self):
        self.scales = (self.bounds[:,1] - self.bounds[:,0])

    def get_bounds(self):
        return self.bounds

    def scale(self, p):
        return (p - self.bounds[:,0])/self.scales * BASINHOPPING_SCALE

    def scale_back(self, p):
        return p/BASINHOPPING_SCALE * self.scales + self.bounds[:,0]

class MonoporeParamsScaler(ParamsScaler):
    def __init__(self, egh_moments_list, num_scales, peak_rgs, unreliable_indeces, num_timescales=0, allow_near_t0=False):
        self.num_scales = num_scales
        self.num_unreliables = len(unreliable_indeces)
        eghM = egh_moments_list[0]
        if allow_near_t0:
            t0_upper = eghM[0]
        else:
            t0_upper = eghM[0] - 5*np.sqrt(eghM[1])
        reliable_indeces = np.setdiff1d(np.arange(len(peak_rgs)), unreliable_indeces)
        reliable_rgmax = np.max(peak_rgs[reliable_indeces])
        self.bounds = make_monopore_bounds(t0_upper, num_scales, self.num_unreliables, reliable_rgmax, num_timescales)
        self.set_scales()

class OligoporeParamsScaler(ParamsScaler):
    def __init__(self, num_pszs, egh_moments_list, num_scales, peak_rgs, unreliable_indeces=[], num_timescales=0, dispersive=False, allow_near_t0=False, t0_lower=None):
        self.num_scales = num_scales
        self.num_unreliables = len(unreliable_indeces)
        eghM = egh_moments_list[0]
        if allow_near_t0:
            t0_upper = eghM[0]
        else:
            t0_upper = eghM[0] - 5*np.sqrt(eghM[1])
        reliable_indeces = np.setdiff1d(np.arange(len(peak_rgs)), unreliable_indeces)
        reliable_rgmax = np.max(peak_rgs[reliable_indeces])
        self.bounds = make_oligopore_bounds(t0_lower, t0_upper, num_pszs, num_scales, self.num_unreliables, reliable_rgmax, num_timescales, dispersive=dispersive)
        self.set_scales()


# coding: utf-8
"""
    QmmController.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from .PairedDataSets import PairedDataSets, get_denoise_rank_impl, get_num_components
from .QuadMM import QuadMM
from .QmmDialog import QmmDialog

class QmmController:
    def __init__(self, qmm, in_folder, sd, separately=True, exe_queue=None):
        self.qmm = qmm
        self.logger = logging.getLogger()
        self.dialog = None
        self.dialog_error = None
        self.exe_queue = exe_queue
        try:
            self.prepare_datasets(in_folder, sd, separately)
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            self.logger.error(etb)
            raise RuntimeError("Error in QmmController.prepare_datasets", etb)

    def put_step_info(self, stepno):
        if self.exe_queue is not None:
            print('put_step_info', [0, stepno])
            self.exe_queue.put([0, stepno])

    def prepare_datasets(self, in_folder, sd, separately):
        self.put_step_info(1)
        kwargs = {}
        kwargs['sd'] = sd
        kwargs['lpm_correct'] = True
        n_components = 10
        kwargs['n_components'] = n_components
        self.datasets = datasets = PairedDataSets(in_folder, kwargs)

        ecurve = datasets.pair[1].ecurve
        denoise_rank = get_denoise_rank_impl(ecurve)
        set_setting('last_denoise_rank', denoise_rank)

        self.put_step_info(2)
        y_list, data_list = datasets.generate_sample_datasets(quad=True, bubble_care=True)
        self.put_step_info(3)

        X_list = [np.expand_dims(data,1) for data in data_list]
        bins_list = [len(y) for y in y_list]
        if separately:
            self.qmm.separate_fit(X=X_list, bins=bins_list)
        else:
            self.qmm.unified_fit(X=X_list, bins=bins_list)
        self.y_list = y_list
        self.data_list = data_list

        self.put_step_info(4)

    def prepare_dialog(self, parent):
        # must be called in the main thread
        try:
            self.dialog = QmmDialog(parent, self)
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            self.logger.error(etb)
            self.dialog_error = etb
            raise RuntimeError("Error in QmmController.show_dialog", etb)      

    def show_dialog(self, parent):
        self.dialog.show()
        return self.dialog.applied

    def dialog_ready(self):
        if self.dialog_error is None:
            return self.dialog is not None
        else:
            # avoid the above waiting forever in test
            raise RuntimeError("Error in QmmController.show_dialog", self.dialog_error)

def get_qmm_controller(parent, mm_type, in_folder, n_components=None, max_iterations=None, sd=None):
    from ProgressMinDialog import run_with_progress

    """
    start loggin copied from Analyzer.do_analysis
    """
    if parent.testing:
        tester_info_log = ' with test pattern ' + str( parent.tester_info.test_pattern )
    else:
        tester_info_log = ''

    parent.analyzer.app_logger.info( "start analysis for " + in_folder + tester_info_log )

    if n_components is None:
        auto_n_components = get_setting('auto_n_components')
        if auto_n_components:
            if sd is None:
                n_components = 10
            else:
                n_components = get_num_components(sd.get_xray_curve())
        else:
            n_components = get_setting('n_components')

    if max_iterations is None:
        max_iterations = get_setting('max_iterations')

    qmm_separately = get_setting('qmm_separately')
    fixed_random_seeds = get_setting('fixed_random_seeds')

    def get_controller(exe_queue):
        if fixed_random_seeds:
            states = fixed_random_seeds
        else:
            states = np.random.randint(1000, 9999, 4)
        set_setting('last_random_seeds', states)
        qmm = QuadMM(mm_type, n_components, max_iter=max_iterations, random_states=states, anim_data=True)
        controller = QmmController(qmm, in_folder, sd, qmm_separately, exe_queue)
        return controller

    def on_return(ret):
        # ret: controller
        ret.prepare_dialog(parent)  # must be called in the main thread

    try:
        controller, error_info = run_with_progress(parent, get_controller, max_iter=5,
                                    title="QmmController", on_return=on_return)
    except:
        controller = None
    return controller



"""
    SecTheory.cf2DistFFT.py

    Translated from MATLAB code by Viktor Witkovsky
    at https://github.com/witkovsky/CharFunTool/CF_InvAlgorithms/cf2DistFFT.m

    Copyright (c) 2022, SAXS Team, KEK-PF

%cf2DistFFT(cf,x,prob,options) Evaluates the CDF/PDF/QF (quantiles) from
% the characteristic function CF of a (continuous) distribution F by using
% the Fast Fourier Transform (FFT) algorithm. 
%
% The algorithm cf2DistFFT evaluates the approximate values CDF(x), PDF(x),
% and/or the quantiles QF(prob) for given x and prob, by interpolation from
% the PDF-estimate computed by the numerical inversion of the given
% characteristic function CF by using the FFT algorithm.
%
% SYNTAX:
%  result = cf2DistFFT(cf,x,prob,options)
%  [result,cdf,pdf,qf] = cf2DistFFT(cf,x,prob,options)
%
% INPUT:
%  cf      - function handle for the characteristic function CF,
%  x       - vector of values from domain of the distribution F, if x = [],
%            cf2DistFFT automatically selects vector x from the domain.
%  prob    - vector of values from [0,1] for which the quantiles will be
%            estimated, if prob = [], cf2DistFFT automatically selects
%            vector prob = [0.9,0.95,0.975,0.99,0.995,0.999].
%  options - structure with the following default parameters:   
%            options.isCompound = false  % treat the compound distributions
%                                        % of the RV Y = X_1 + ... + X_N,
%                                        % where N is discrete RV and X>=0 
%                                        % are iid RVs from nonnegative
%                                        % continuous distribution.
%             options.isInterp   = false % create and use the interpolant
%                                          functions for PDF/CDF/QF/RND
%             options.N = 2^10         % N points used by FFT
%             options.xMin = -Inf      % set the lower limit of X
%             options.xMax = Inf       % set the lower limit of X
%             options.xMean = []       % set the MEAN value of X
%             options.xStd = []        % set the STD value of X
%             options.dt = []          % set grid step dt = 2*pi/xRange
%             options.T = []           % set upper limit of (0,T), T = N*dt 
%             options.SixSigmaRule = 6 % set the rule for computing domain
%             options.tolDiff = 1e-4   % tol for numerical differentiation
%             options.isPlot = true    % plot the graphs of PDF/CDF
%  options.DIST - structure with information for future evaluations.
%                 options.DIST is created automatically after first call: 
%             options.DIST.xMin  = xMin   % the lower limit of X
%             options.DIST.xMax  = xMax   % the upper limit of X
%             options.DIST.xMean = xMean  % the MEAN value of X, 
%             options.DIST.cft   = cft    % CF evaluated at t_j : cf(t_j).   
%
% OUTPUTS:
%  result  - structure with the following results values:  
%             result.x                  = x;
%             result.cdf                = cdf;
%             result.pdf                = pdf;
%             result.prob               = prob;
%             result.qf                 = qf;
%             result.xFTT               = xFFT;
%             result.pdfFFT             = pdfFFT;
%             result.cdfFFT             = cdfFFT;
%             result.SixSigmaRule       = options.SixSigmaRule;
%             result.N                  = N;
%             result.dt                 = dt;
%             result.T                  = t(end);
%             result.PrecisionCrit      = PrecisionCrit;
%             result.myPrecisionCrit    = options.crit;
%             result.isPrecisionOK      = isPrecisionOK;
%             result.xMean              = xMean;
%             result.xStd               = xStd;
%             result.xMin               = xMin;
%             result.xMax               = xMax;
%             result.cf                 = cf;
%             result.options            = options;
%             result.tictoc             = toc;
%
% EXAMPLE 1:
%  % DISTRIBUTION OF A LINEAR COMBINATION OF THE INDEPENDENT RVs
%  % (Normal, Student's t, Rectangular, Triangular & Arcsine distribution)
%  % Y = X_{N} + X_{t} + 5*X_{R} + X_{T} + 10*X_{U}
%  % CFs: Normal, Student's t, Rectangular, Triangular, and Arcsine
%  cf_N  = @(t) exp(-t.^2/2);                                      
%  cf_t  = @(t,nu) min(1,besselk(nu/2, abs(t).*sqrt(nu),1) .* ...
%          exp(-abs(t).*sqrt(nu)) .* (sqrt(nu).*abs(t)).^(nu/2) / ...
%          2^(nu/2-1)/gamma(nu/2));     
%  cf_R  = @(t) min(1,sin(t)./t);   
%  cf_T  = @(t) min(1,(2-2*cos(t))./t.^2);
%  cf_U  = @(t) besselj(0,t);  
%  % Characteristic function of the linear combination Y
%  c    = [1 1 5 1 10]; nu = 1;
%  cf_Y = @(t) cf_N(c(1)*t) .* cf_t(c(2)*t,nu) .* cf_R(c(3)*t) .* ...
%         cf_T(c(4)*t) .* cf_U(c(5)*t);
%  clear options
%  options.N    = 2^10;
%  options.xMin = -50;
%  options.xMax = 50;
%  result = cf2DistFFT(cf_Y,[],[],options);
%  title('CDF of Y = X_{N} + X_{t} + 5*X_{R} + X_{T} + 10*X_{U}')
%
% EXAMPLE 2:
%  % DISTRIBUTION OF A LINEAR COMBINATION OF THE INDEPENDENT CHI2 RVs
%  % (Chi-squared RVs with 1 and 10 degrees of freedom)
%  % Y = 10*X_{Chi2_1} + X_{Chi2_10}
%  % Characteristic functions of X_{Chi2_1} and X_{Chi2_10}
%  %
%  df1       = 1;
%  df2       = 10;
%  cfChi2_1  = @(t) (1-2i*t).^(-df1/2);
%  cfChi2_10 = @(t) (1-2i*t).^(-df2/2);
%  cf_Y      = @(t) cfChi2_1(10*t) .* cfChi2_10(t);
%  clear options
%  options.xMin = 0;
%  result = cf2DistFFT(cf_Y,[],[],options);
%  title('CDF of Y = 10*X_{\chi^2_{1}} + X_{\chi^2_{10}}')
%
% EXAMPLE3 (PDF/CDF of the compound Poisson-Exponential distribution)
%  lambda1 = 10;
%  lambda2 = 5;
%  cfX  = @(t) cfX_Exponential(t,lambda2);
%  cf   = @(t) cfN_Poisson(t,lambda1,cfX);
%  x    = linspace(0,8,101);
%  prob = [0.9 0.95 0.99];
%  clear options
%  options.isCompound = 1;
%  result = cf2DistFFT(cf,x,prob,options)
%
% REMARKS:
%  The outputs of the algorithm cf2DistFFT are approximate values! The
%  precission of the presented results depends on several different
%  factors: 
%  - application of the FFT algorithm for numerical inversion of the CF
%  - selected number of points used by the FFT algorithm (by default
%    options.N = 2^10),
%  - estimated/calculated domain [A,B] of the distribution F, used with the
%    FFT algorithm. Optimally, [A,B] covers large part of the
%    distribution domain, say more than 99%. However, the default
%    automatic procedure for selection of the domain [A,B] may fail. It
%    is based on the 'SixSigmaRule': A = MEAN - SixSigmaRule * STD, and
%    B = MEAN + SixSigmaRule * STD. Alternatively, change the
%    options.SixSigmaRule to different value, say 12, or use the
%    options.xMin and options.xMax to set manually the values of A and B.
%
% REFERENCES:
% [1] WITKOVSKY, V.: On the exact computation of the density and of
%     the quantiles of linear combinations of t and F random
%     variables. Journal of Statistical Planning and Inference 94
%     (2001), 1-13.
% [2] WITKOVSKY, V.: Matlab algorithm TDIST: The distribution of a
%     linear combination of Student's t random variables. In COMPSTAT
%     2004 Symposium (2004), J. Antoch, Ed., Physica-Verlag/Springer
%     2004, Heidelberg, Germany, pp. 1995-2002.
% [3] WITKOVSKY, V.: WIMMER,G., DUBY, T. Logarithmic Lambert W x F
%     random variables for the family of chi-squared distributions
%     and their applications. Statistics & Probability Letters 96
%     (2015), 223-231.  
% [4] WITKOVSKY V. (2016). Numerical inversion of a characteristic
%     function: An alternative tool to form the probability distribution of
%     output quantity in linear measurement models. Acta IMEKO, 5(3), 32-44.
%
% SEE ALSO: cf2Dist, cf2DistGP, cf2DistGPT, cf2DistGPA, cf2DistFFT,
%           cf2DistBV, cf2CDF, cf2PDF, cf2QF 

% (c) Viktor Witkovsky (witkovsky@gmail.com)
% Ver.: 01-Sep-2020 13:25:21
%
% Revision history:
% Ver.: 15-Nov-2016 13:36:26
"""
import numpy as np

class Options:
    pass

class DistInfo:
    pass

def cf2DistFFT(cf, x=None, prob=None, options=Options()):

    if not hasattr(options, 'isCompound'):
        options.isCompound = False

    if not hasattr(options, 'N'):
        if options.isCompound:
            options.N = 2**14
        else:
            options.N = 2**10

    if not hasattr(options, 'xMin'):
        if options.isCompound:
            options.xMin = 0
        else:
            options.xMin = -np.inf

    if not hasattr(options, 'xMax'):
        options.xMax = np.inf

    if not hasattr(options, 'xMean'):
        options.xMean = None

    if not hasattr(options, 'xStd'):
        options.xStd = None

    if not hasattr(options, 'dt'):
        options.dt = None

    if not hasattr(options, 'T'):
        options.T = None

    if not hasattr(options, 'SixSigmaRule'):
        if options.isCompound:
            options.SixSigmaRule = 10
        else:
            options.SixSigmaRule = 6

    if not hasattr(options, 'tolDiff'):
        options.tolDiff = 1e-4

    if not hasattr(options, 'crit'):
        options.crit = 1e-12;

    if not hasattr(options, 'isPlot'):
        options.isPlot = True

    if not hasattr(options, 'DIST'):
        options.DIST = DistInfo()

    # Other options parameters
    if not hasattr(options, 'isPlotFFT'):
        options.isPlotFFT = False;

    if not hasattr(options, 'xN'):
        options.xN = 101;

    if not hasattr(options, 'chebyPts'):
        options.chebyPts = 2**9

    if not hasattr(options, 'isInterp'):
        options.isInterp = False

    ## GET/SET the DEFAULT parameters and the OPTIONS
    # First, set a special treatment if the real value of CF at infinity (large
    # value) is positive, i.e. const = real(cf(Inf)) > 0. In this case the
    # compound CDF has jump at 0 of size equal to this value, i.e. cdf(0) =
    # const, and pdf(0) = Inf. In order to simplify the calculations, here we
    # calculate PDF and CDF of a distribution given by transformed CF, i.e.
    # cf_new(t) = (cf(t)-const) / (1-const); which is converging to 0 at Inf,
    # i.e. cf_new(Inf) = 0. Using the transformed CF requires subsequent
    # recalculation of the computed CDF and PDF, in order to get the originaly
    # required values: Set pdf_original(0) =  Inf & pdf_original(x) =
    # pdf_new(x) * (1-const), for x > 0. Set cdf_original(x) =  const +
    # cdf_new(x) * (1-const).
    # 

    const = cf(1e30).real
    if options.isCompound:
        cfOld = cf
        if const > 1e-13:
            cf = lambda t: (cf(t) - const) / (1-const)

    if hasattr(options.DIST, 'xMean'):
        xMean              = options.DIST.xMean
        cft                = options.DIST.cft
        xMin               = options.DIST.xMin
        xMax               = options.DIST.xMax
        N                  = len(cft)
        k                  = np.arange(N)   # (0:(N-1))';
        xRange             = xMax - xMin
        dt                 = 2*np.pi / xRange
        t                  = (k - N/2 + 0.5) * dt
        xStd               = None
    else:
        N                  = 2*options.N
        dt                 = options.dt
        T                  = options.T
        xMin               = options.xMin
        xMax               = options.xMax
        xMean              = options.xMean
        xStd               = options.xStd
        SixSigmaRule       = options.SixSigmaRule
        tolDiff            = options.tolDiff
        cft                = cf(tolDiff*np.arange(1,5))     # (1:4)

        if xMean is None:
            xMean = ((-cft[1]
                + 8*cft[0]-8*cft[0].conj()
                + cft[1].conj())/(1j*12*tolDiff)).real

        if xStd is None:
            xM2 = (-(cft[3].conj()
                - 16*cft[2].conj()
                + 64*cft[1].conj()
                + 16*cft[0].conj()
                - 130 + 16*cft[0]
                + 64*cft[1]
                - 16*cft[2]+cft[3])/(144*tolDiff**2)).real
            xStd  = np.sqrt(xM2 - xMean**2)

        if np.isfinite(xMin) and np.isfinite(xMax):
            xRange = xMax - xMin
        elif T is not None:
            xRange = 2*np.pi / (2 * T / N)
            if np.isfinite(xMax):
                xMin = xMax - xRange
            elif np.isfinite(xMin):
                xMax = xMin + xRange
            else:
                xMin = xMean - xRange/2
                xMax = xMean + xRange/2
        elif dt is not None:
            xRange = 2*np.pi / dt
            if np.isfinite(xMax):
                xMin = xMax - xRange
            elif np.isfinite(xMin):
                xMax = xMin + xRange
            else:
                xMin = xMean - xRange/2
                xMax = xMean + xRange/2
        else:
            if np.isfinite(xMin):
                xMax = xMean + SixSigmaRule * xStd
            elif np.isfinite(xMax):
                xMin = xMean - SixSigmaRule * xStd
            else:
                xMin = xMean - SixSigmaRule * xStd
                xMax = xMean + SixSigmaRule * xStd
            xRange = xMax - xMin

        dt                 = 2*np.pi / xRange
        k                  = np.arange(N, dtype=complex)    # (0:(N-1))'
        t                  = (k - N/2 + 0.5) * dt
        cft                = cf(t[N//2:])
        cft                = np.concatenate([cft[::-1].conj(), cft])
        # cft[0]             = cft[0]/2
        # cft[N-1]           = cft[N-1]/2
        options.DIST.xMin  = xMin
        options.DIST.xMax  = xMax
        options.DIST.xMean = xMean
        options.DIST.cft   = cft

    ## ALGORITHM

    A      = xMin
    B      = xMax
    dx     = (B-A)/N
    c      = (-1)**(A*(N-1)/(B-A))/(B-A)
    print("A, B, N, dx, c=", A, B, N, dx, c)
    C      = c * (-1)**((1-1/N)*k)
    D      = (-1)**(-2*(A/(B-A))*k)     # k must be complex, see https://stackoverflow.com/questions/45384602/numpy-runtimewarning-invalid-value-encountered-in-power
    pdfFFT = np.max([np.zeros(N), (C*np.fft.fft(D*cft)).real], axis=0)

    # Reset the transformed CF, PDF, and CDF to the original values
    if options.isCompound:
        pdfFFT = pdfFFT * (1-const)
        pdfFFT[x==0] = np.inf

    return pdfFFT

if __name__ == '__main__':
    from scipy.special import jv, kve, gamma
    import matplotlib.pyplot as plt

    def ones(x):
        return 1 if np.isscalar(x) else np.ones(len(x))

    cf_N = lambda t: np.exp(-t**2/2)
    cf_t = lambda t, nu: np.min([ones(t), kve(nu/2, np.abs(t)*np.sqrt(nu)) * np.exp(-np.abs(t)* np.sqrt(nu)) * (np.sqrt(nu)*np.abs(t))**(nu/2) / 2**(nu/2-1) / gamma(nu/2)], axis=0)
    cf_R = lambda t: np.min([ones(t), np.sin(t)/t], axis=0)
    cf_T = lambda t: np.min([ones(t), (2-2*np.cos(t))/t**2], axis=0)
    cf_U = lambda t: jv(0, t)

    # Characteristic function of the linear combination Y
    c = np.array([1, 1, 5, 1, 10])
    nu = 1
    cf_Y = lambda t: cf_N(c[0]*t) * cf_t(c[1]*t,nu) * cf_R(c[2]*t) * cf_T(c[3]*t) * cf_U(c[4]*t)

    options = Options()
    options.N    = 2**10
    options.xMin = -50
    options.xMax = 50

    pdf = cf2DistFFT(cf_Y, None, None, options)

    print("pdf=", pdf[0:10])
    plt.plot(pdf)
    plt.show()



# coding: utf-8
"""
    OurStatsModels.py

    Copyright (c) 2016-2018, Masatsuyo Takahashi, KEK-PF
"""
import numpy            as np

def add_constant( x ):
    return np.vstack( [ np.ones( ( len(x), ) ), x ] ).T

class Result:
    def __init__( self, **entries ): 
        self.__dict__.update(entries)

    def cov_params( self ):
        return self.cov_params_

class WLS:
    def __init__( self, y, X, weights=None ):
        if type(y) != np.ndarray:
            y = np.array(y)
        if weights is None:
            weights = np.ones( ( len(y), ) )
        else:
            if type(weights) != np.ndarray:
                weights = np.array(weights)

        assert( len( y.shape ) == 1 )
        assert( len( X.shape ) == 2 )
        assert( len( weights.shape ) == 1 )
        assert( y.shape[0] == X.shape[0] )
        assert( y.shape[0] == weights.shape[0] )

        self.y  = y
        self.A          = X
        self.N  = N = y.shape[0]
        shape_ = ( N, 1 )
        # print( self.A )
        self.y_         = np.reshape( y, shape_ )
        self.w          = weights
        self.W          = np.diag( weights )

    def fit( self ):
        At      = np.transpose( self.A )
        AtW     = np.dot( At, self.W )
        self.XtWX_inv = AtWA_inv = np.linalg.inv( np.dot( AtW, self.A ) )
        self.XtWX_inv_XtW = AtWX_inv_AtW = np.dot( AtWA_inv, AtW )
        params  = np.dot( AtWX_inv_AtW, self.y_ )
        # print( params )
        """
            As for the covariance matrix of params, see
            Linear least squares (mathematics)
            https://en.wikipedia.org/wiki/Linear_least_squares_%28mathematics%29#Weighted_linear_least_squares
        """
        e       = ( self.y_ - np.dot( self.A, params ) ).squeeze()
        # print( 'e_=', e[0:5] )
        # print( 'e =', 1/np.sqrt( self.w[0:5] ) )
        ee = e**2
        S       = np.inner( self.w,  ee)
        sigma2  = S / ( self.N - 2 )
        # print( 'S=', S, ', N=', self.N, ', sigma2=', sigma2 )
        cov_params = AtWA_inv * sigma2

        ssr = np.sum(ee)
        # TODO: this ssr value is differenct from that of statsmodels.api

        return Result(
                params=np.reshape( params, ( params.shape[0], ) ),
                cov_params_=cov_params,
                ssr=ssr,
                )

class OLS(WLS):
    def __init__(self, y, X):
        WLS.__init__(self, y, X)



"""
    update-denss.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import sys
import os
from shutil import copy
import diff_match_patch as dmp_module

try:
    GIT_REPOSITORY = r"D:\Github\denss"
    assert os.path.exists(GIT_REPOSITORY)
except:
    GIT_REPOSITORY = r"E:\Github\denss"
    assert os.path.exists(GIT_REPOSITORY)

denss_scripts = os.path.join(GIT_REPOSITORY, r"denss\scripts")
denss_lib = os.path.join(GIT_REPOSITORY, "denss")

temp_scripts = "scripts-temp"
temp_lib = "core-temp"

for path in [temp_scripts, temp_lib]:
    if not os.path.exists(path):
        os.makedirs(path)

this_dir = os.path.dirname(os.path.abspath( __file__ ))
root_dir = os.path.dirname(os.path.dirname(this_dir))
sys.path.insert(0, root_dir)

from molass_legacy.KekLib.DiffUtils import file2string, string2file

dmp = dmp_module.diff_match_patch()
dmp.Match_Distance = 5000

# bin
old_bin = "bin"

for file in ["denss_fit_data.py", "denss_pdb2mrc.py"]:
    orig_file = file.replace(".py", "-orig.py")

    old_src = file2string(os.path.join(old_bin, orig_file))
    mod_src = file2string(os.path.join(old_bin, file))

    patches = dmp.patch_make(old_src, mod_src)
    print(dmp.patch_toText(patches))

    git_src_path = os.path.join(denss_scripts, file)
    git_src = file2string(git_src_path)
    results = dmp.patch_apply(patches, git_src)

    new_src = results[0]
    print("=====================", results[1])
    assert results[1] == [True] * len(results[1])

    string2file(new_src, os.path.join(temp_scripts, file))
    copy(git_src_path, os.path.join(temp_scripts, orig_file))

exit()

# saxstats
old_lib = "saxstats"

for file in ["denssopts.py", "saxstats.py"]:
    orig_file = file.replace(".py", "-orig.py")

    old_src = file2string(os.path.join(old_lib, orig_file))
    mod_src = file2string(os.path.join(old_lib, file))

    patches = dmp.patch_make(old_src, mod_src)
    print(dmp.patch_toText(patches))

    git_src_path = os.path.join(git_lib, file)
    git_src = file2string(git_src_path)
    results = dmp.patch_apply(patches, git_src)

    new_src = results[0]
    print("=====================", results[1])
    assert results[1] == [True] * len(results[1])

    string2file(new_src, os.path.join(tmp_lib, file))
    copy(git_src_path, os.path.join(tmp_lib, orig_file))

for file in ["__init__.py", "_version.py"]:
    git_src_path = os.path.join(git_lib, file)
    copy(git_src_path, os.path.join(tmp_lib, file))
    print("copied", file)



# coding: utf-8
"""

    ファイル名：   MappingZoomCanvas.py

    処理内容：

        濃度対応図の表示

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF

"""
import matplotlib.pyplot                as plt
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter                         import Tk, Dialog
from molass_legacy.KekLib.TkUtils                            import split_geometry, is_low_resolution
from molass_legacy.KekLib.OurMatplotlib                      import NavigationToolbar

DEBUG = False

class MappingZoomCanvas( Tk.Toplevel ):
    def __init__( self, parent, draw_closure, annotation_closure=None, quit_cb=None ):
        self.parent         = parent
        self.draw_closure   = draw_closure
        self.annotation_closure = annotation_closure
        self.quit_cb        = quit_cb
        Tk.Toplevel.__init__( self, parent )
        self.create_canvas()
        self.update()

        W, H, X, Y = split_geometry( self.parent.geometry() )
        self.geometry("+%d+%d" % ( X + 50, Y + 50 ))

        self.protocol( "WM_DELETE_WINDOW", self.quit )

    def quit( self ):
        if self.quit_cb is not None:
            self.quit_cb()
        self.destroy()

    def create_canvas( self ):
        self.cframe = cframe = Tk.Frame( self )
        cframe.pack()

        figsize = ( 10, 8 ) if is_low_resolution() else ( 12, 10 )
        self.fig = fig = plt.figure( figsize=figsize )

        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )

        self.ax = fig.add_subplot(111)
        self.draw()
        self.fig.tight_layout()

        self.toolbar = NavigationToolbar( self.mpl_canvas, cframe )
        self.toolbar.update()

        close_button = Tk.Button( self, text="Close", command=self.quit )
        close_button.pack( pady=5 )

    def draw( self, draw_closure=None ):
        if draw_closure is not None:
            self.draw_closure = draw_closure
        self.draw_closure( self.ax )
        if self.annotation_closure is not None:
            self.annotation_closure( self.ax, zoomed=True )
        self.mpl_canvas.draw()

    def canvas_focus_set( self ):
        self.mpl_canvas_widget.focus_set()



"""
    Solvers.BH.SolverBH.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from .BasinHopping import BasinHopping

NARROW_BIND_ALLOW = 1.0

class SolverBH:
    def __init__(self, optimizer):
        self.bh_impl = BasinHopping()
        self.optimizer = optimizer

    def minimize(self, objective, init_params, niter=100, seed=1234, bounds=None, narrow_bounds=False, show_history=False):

        if narrow_bounds:
            lower = init_params - NARROW_BIND_ALLOW
            upper = init_params + NARROW_BIND_ALLOW
            bounds = np.array([lower, upper]).T

        minimizer_kwargs = dict(method='Nelder-Mead', bounds=bounds)

        minima_callback = self.optimizer.minima_callback
        accept_test = self.optimizer.accept_test

        result = self.bh_impl.minimize(objective, init_params, niter=niter, seed=seed,
                                callback=minima_callback,
                                accept_test=accept_test,
                                minimizer_kwargs=minimizer_kwargs,
                                custom=show_history)

        if show_history:
            self.bh_impl.show_history(self)

        return result
    
    def show_history(self, caller):
        self.bh_impl.show_history(caller)


"""
    find_the_most_likely.py

"""
import sys
sys.path += [
    r"D:\PyTools\pytools-2_9_5_develop\lib",
    r"D:\PyTools\pytools-2_9_5_develop\lib\KekLib",
    r"D:\PyTools\pytools-2_9_5_develop\lib\DataStructure",
    r"D:\PyTools\pytools-2_9_5_develop\lib\SerialAnalyzer",
    ]

import os
import numpy as np
from scipy.interpolate import UnivariateSpline
import matplotlib.pyplot as plt
import seaborn
seaborn.set()
from glob import glob
import subprocess
from CrysolUtils import np_loadtxt_crysol
from molass_legacy.DENSS.DenssUtils import fit_data

debug = False

this_dir = os.path.dirname(os.path.abspath(__file__))
crysol_dir = os.path.join(this_dir, "crysol_files")

component_file = r"D:\TODO\20231031\CRYSOL\component-1.dat"
crysol_exe = r"C:\Program Files\ATSAS-3.1.3\bin\crysol.exe"

target = np.loadtxt(component_file)
print(target.shape)
qc, Ic, Icerr, D = fit_data(*target.T)

qv = None
ty = None
target_spline = UnivariateSpline(qc, Ic, s=0, ext=3)

def normalize(y):
    i = len(y)//4
    return y/np.max(y[0:i])

def compute_rmsd(y1, y2):
    return np.sqrt(np.mean((y1 - y2)**2))

min_rmsd = None
min_code = None

fh = open("rmsd.csv", "w")

for k, file in enumerate(glob(r"D:\TODO\20231031\PDB\aldolase_non_mutants\*.*")):
    print([k], file)
    if file[-3:] == "pdb":
        pass
    else:
        new_name = file[:-5] + "-ba" + file[-1] + ".pdb"
        os.rename(file, new_name)
        file = new_name

    crysol_file = file[-12:-3] + "int"
    crysol_path = os.path.join(crysol_dir, crysol_file)
    if os.path.exists(crysol_path):
        pass
    else:
        # uncomment this to stop adding crysol results
        # continue

        if os.path.exists(crysol_file):
            os.rename(crysol_file, crysol_path)
        else:
            cmd = [crysol_exe, file, component_file, '--constant']
            result = subprocess.run(cmd, capture_output=True)
            if result.returncode == 0:
                os.rename(crysol_file, crysol_path)
            else:
                print(result.stderr)
                continue

    # cleanup
    print(crysol_file)
    pdbcode = crysol_file[:-4]
    for file in glob(pdbcode + "*.*"):
        print("delete", file)
        os.remove(file)

    crysol_data = np_loadtxt_crysol(crysol_path)[0]

    if qv is None:
        qv = crysol_data[:,0]
        ty = normalize(target_spline(qv))

    cy = normalize(crysol_data[:,1])
    rmsd = compute_rmsd(cy, ty)
    fh.write(",".join([pdbcode, "%g" % rmsd]) + "\n")
    fh.flush()
    if min_rmsd is None or rmsd < min_rmsd:
        min_rmsd = rmsd
        min_code = pdbcode
    if debug:
        print("rmsd=", rmsd)
        fig, ax = plt.subplots()
        ax.set_yscale('log')
        ax.plot(qv, ty)
        ax.plot(qv, cy)
        fig.tight_layout()
        plt.show()

    if k > 1:
        # break
        pass

print("min_code=%s, min_rmsd=%g" % (min_code, min_rmsd))

fh.close()



"""
    BoundedLrfSolver.py

    Copyright (c) 2023-2025, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from bisect import bisect_right
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.DataStructure.SvdDenoise import get_denoised_data
from molass_legacy.KekLib.SciPyCookbook import smooth
from molass_legacy.LRF.LrfRgUtils import compute_rg_from_qvDEP
from molass_legacy.Theory.SolidSphere import phi
from .ErrorCorrection import compute_corrected_error

class BoundedLrfSolver:
    def __init__(self, qv, D, E, C=None, i=None, denoise=True, debug_info=None):
        if i is None:
            from bisect import bisect_right
            i = bisect_right(qv, 0.02)

        self.debug_info = debug_info
        self.qv = qv
        self.D = D
        self.i = i
        if C is None:
            cy = np.average(D_[max(0, i-5):i+6,:], axis=0)      # i=3 with 20200304_1
            self.Cinit = np.array([cy, cy**2])
        else:
            cy = C[0,:]
            self.Cinit = C
        self.j = np.argmax(cy)

        if denoise:
            D_ = get_denoised_data(D, rank=C.shape[0])
        else:
            D_ = D

        self.D_ = D_
        self.E = E

    def solve(self, debug=False):
        qv = self.qv
        i = self.i
        j = self.j
        C = self.Cinit
        D_ = self.D_
        try:
            Cinv = np.linalg.pinv(C)
        except:
            cx = np.arange(C.shape[1])
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("solve pinv debug")
                for k, cy in enumerate(C):
                    print([k], cy)
                    ax.plot(cx, cy)
                fig.tight_layout()
                plt.show()
        P = D_ @ Cinv
        c1, c2 = C[0:2,j]

        Rg, R, L_, hK, hL = self.estimate_RgL(P, c1, debug=debug)

        aq, bq = P.T[0:2]

        # P_ = P.copy()
        normD = np.linalg.norm(self.D_)
        normBq = np.linalg.norm(bq)
        P_ = P.copy()

        def coerce_bounds(L_, return_info=False):
            bound = 1/(qv*L_*R)**2

            bq_bound = bound * aq/c1
            bq_bounds = (-bq_bound, bq_bound)

            coerced_bq = P_[:,1]
            where = bq < bq_bounds[0]
            coerced_bq[where] = bq_bounds[0][where]
            where = bq > bq_bounds[1]
            coerced_bq[where] = bq_bounds[1][where]

            P_[:,0] = aq + (bq - coerced_bq)*c2/c1

            if return_info:
                return bq_bounds, coerced_bq, P_

            return np.log10(np.linalg.norm(self.D_ - P_ @ C)/normD) + np.log10(np.linalg.norm(coerced_bq - bq)/normBq)

        bq_bounds, coerced_bq, P_ = coerce_bounds(L_, return_info=True)

        debug_plot = debug and self.debug_info is not None
        if debug_plot:
            from DataUtils import get_in_folder
            print("L_=", L_)
            in_folder = get_in_folder()
            aq_, bq_ = P_.T
            with plt.Dp():
                fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18,5))
                fig.suptitle("Bounded LRF Debug using %s" % in_folder)
                ax2.set_yscale("log")
                for i, ax in enumerate((ax1, ax2)):
                    ax.plot(qv, aq*c1)
                    ax.plot(qv, aq_*c1)
                    if i == 0:
                        ax.plot(qv, bq*c2)
                        ax.plot(qv, bq_*c2)
                ax3.plot(qv, bq*c2)
                ax3.plot(qv, coerced_bq*c2)
                ax3.plot(qv, bq_bounds[0]*c2)
                ax3.plot(qv, bq_bounds[1]*c2)
                ax3.set_ylim(ax1.get_ylim())
                fig.tight_layout()
                plt.show()

        D_pinv = np.linalg.pinv(self.D)
        W   = np.dot(D_pinv, P_)
        Pe  = np.sqrt(np.dot(self.E**2, W**2) )

        self.Pe_, adjust_scale = compute_corrected_error(qv, i, P, c1, c2, P_, Pe, bq_bounds, coerced_bq)

        return P_, C, Rg, R, L_, hK, hL, bq_bounds, coerced_bq

    def get_corrected_error(self):
        return self.Pe_

    def estimate_RgL(self, P, c1, debug=False):
        from scipy.optimize import basinhopping
        qv = self.qv
        aq, bq = P.T[0:2]
        sg = compute_rg_from_qvDEP(qv, self.D_, self.E, P, return_sg=True)
        Rg = sg.Rg

        debug_plot = debug and self.debug_info is not None
        if False:
            if Rg is None:
                glim = 0.02
            else:
                glim = 1.5/Rg
            glim_i = bisect_right(qv, glim)
            gslice = slice(0,glim_i)
            print("gslice=", gslice)
            q = (1.3/Rg)**2
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("estimate_RgL: Guinier Plot $R_g=%.3g$" % Rg, fontsize=20)
                ax.set_ylabel(r"$log_e(I)$")
                ax.set_xlabel(r"$Q^2$")
                ax.plot(qv[gslice]**2, np.log(aq[gslice]))
                ymin, ymax = ax.get_ylim()
                ax.set_ylim(ymin, ymax)
                ax.plot([q, q], [ymin, ymax], color="yellow", label=r"$qR_g=1.3$")
                ax.legend()
                fig.tight_layout()
                plt.show()

        R = np.sqrt(5/3)*Rg
        K = 0.3
        L = 1

        i1 = bisect_right(qv, 0.01)
        i2 = bisect_right(qv, 0.2)
        slice_ = slice(i1,i2)
        qv_ = qv[slice_]
        aq_ = aq[slice_]
        bq_ = bq[slice_]

        def objective(p):
            K_, L_ = p
            h_bq = -K_*phi(qv_, 2*L_*R)*aq_
            negetive_penalty = min(0, K_)**2 + min(0, L_ - 2)**2
            return np.mean((h_bq - bq_)**2) + negetive_penalty

        # ret = minimize(objective, (K, L), method="Nelder-Mead")
        ret = minimize(objective, (K, L))

        if debug_plot:
            print("c1=", c1)
            K, L = ret.x
            print("(1) K, L =", K, L)
            h_bq_ = -K*phi(qv_, 2*L*R)*aq_
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("estimate_RgL debug")
                ax.plot(qv_, bq_, label="Data B(q)")
                ax.plot(qv_, h_bq_, label="Sphere B(q)")
                ax.legend()
                fig.tight_layout()
                plt.show()

        ig = bisect_right(qv, 0.06)
        gslice = slice(0, ig)
        qvg = qv[gslice]
        qvg2 = qvg**2

        if debug_plot:
            from DataUtils import get_in_folder
            K = 0.29
            L = 1.9
            j = self.j
            Icq = self.D_[:,j]
            d_aq = aq/sg.Iz
            h_aq = phi(qv, R)**2

            Hq = h_aq * (1 - K*phi(qv, L*2*R))
            scale = np.max(Hq)/np.max(Icq)
            Iq = Icq * scale

            # h_aq_ = phi(qv_, R)**2
            h_bq = - h_aq * K * phi(qv, L*2*R)

            Iqg = Iq[gslice]
            Hqg = Hq[gslice]

            d_aqg = d_aq[gslice]
            h_aqg = h_aq[gslice]

            if False:

                def guinier_objective(p):
                    K_, L_ = p
                    Hqg_ = h_aqg * (1 - K_*phi(qvg, L_*2*R))
                    return np.sum((np.log(Hqg_) - np.log(Iqg))**2)

                rret = minimize(guinier_objective, (0.3, 2))
                K, L = ret.x
                print("(2) K, L =", K, L)
                h_bq = - h_aq * K * phi(qv, L*2*R)

            q13 = (1.3/Rg)**2

            in_folder = get_in_folder()
            with plt.Dp():
                fig, axes = plt.subplots(nrows=2, ncols=4, figsize=(20,10))
                fig.suptitle("Hard Sphere Model Fitting Demo using %s descending side, $R_g$=%.3g" % (in_folder, Rg), fontsize=20)

                ax0 = axes[0,0]
                ax1 = axes[0,1]
                ax2 = axes[0,2]
                ax3 = axes[0,3]

                ax4 = axes[1,0]
                ax5 = axes[1,1]
                ax6 = axes[1,2]
                ax7 = axes[1,3]

                for ax in [ax0, ax2, ax4, ax6]:
                    ax.set_yscale("log")

                ax0.set_title("Apparent Log View", fontsize=16)
                ax1.set_title("Apparent Guinier View", fontsize=16)
                ax2.set_title("Measured Only View", fontsize=16)
                ax3.set_title("Linear A(q) View", fontsize=16)
                ax4.set_title("Eoii-removed Log View", fontsize=16)
                ax5.set_title("Eoii-removed Guinier View", fontsize=16)
                ax6.set_title("Model Only View", fontsize=16)
                ax7.set_title("Linear B(q) View", fontsize=16)

                ax0.plot(qv, Iq, label="Data I(q)")
                ax0.plot(qv, Hq, label="Sphere I(q)")
                ax0.legend()

                ax1.plot(qvg2, np.log(Iqg), label="Data I(q)")
                ax1.plot(qvg2, np.log(Hqg), label="Sphere I(q)")

                ax2.plot(qv, Iq, label="Data I(q)")
                ax2.plot(qv, d_aq, label="Data A(q)")
                ax2.legend()

                ax3.plot(qv, d_aq, label="Data A(q)")
                ax3.plot(qv, h_aq, label="Sphere A(q)")
                ax3.legend()

                ax4.plot(qv, d_aq, label="Data A(q)")
                ax4.plot(qv, h_aq, label="Sphere A(q)")
                ax4.legend()

                ax5.plot(qvg2, np.log(d_aqg), label="Data A(q)")
                ax5.plot(qvg2, np.log(h_aqg), label="Sphere A(q)")

                def plot_q13(ax):
                    ymin, ymax = ax.get_ylim()
                    ax.set_ylim(ymin, ymax)
                    ax.plot([q13, q13], [ymin, ymax], color="yellow", label="$qR_g=1.3$")

                for ax in ax1, ax5:
                    plot_q13(ax)
                    ax.legend()

                ax6.plot(qv, Hq, label="Sphere I(q)")
                ax6.plot(qv, h_aq, label="Sphere A(q)")
                ax6.legend()

                ax7.plot(qv, bq, label="Data B(q)")
                h_bq_scaled = h_bq*scale**2
                ax7.plot(qv, h_bq_scaled, label="Sphere B(q)")

                zeros = np.zeros(len(qv))

                def bound_objective(p):
                    L = p[0]
                    return np.sum(np.max([zeros, np.abs(h_bq_scaled) - 1/(qv*L*R)**2], axis=0)**2) - L

                retb = minimize(bound_objective, (1,))
                L = retb.x[0]
                print("L=", L)

                by = 1/(qv*L*R)**2
                ax7.plot(qv, by, ":", color="red", label="B(q) bound")
                ax7.plot(qv, -by,":", color="red")

                ymin, ymax = ax7.get_ylim()
                ax7.set_ylim(-10, 10)
                ax7.legend()
                fig.tight_layout()
                plt.show()

        j2 =  np.argmax(smooth(bq_))
        slice2_ = slice(j2,i2)
        q_ = qv[slice2_]
        aq2_ = aq[slice2_]
        bq2_ = bq[slice2_]

        hK, hL = ret.x
        h_bq = -hK*phi(q_, 2*hL*R)*aq2_/c1
        upper_bq2 = np.max([bq2_, h_bq], axis=0)

        def objective2(p):
            L, = p
            bound = 1/(q_*L*R)**2 * aq2_/c1
            diff = upper_bq2 - bound
            return 99*np.mean(diff[diff>0]**2) + np.mean(diff[diff<=0]**2)

        ret2 = minimize(objective2, (hL,))
        retL = ret2.x[0]

        if debug_plot:
            K, L = ret.x
            print("L=", L, "K=", K)
            print("ret.nit=", ret.nit)
            print("ret2.x[0]=", ret2.x[0])
            h_bq = -K*phi(qv, 2*L*R)*aq/c1
            bound = 1/(qv*retL*R)**2 * aq/c1
            c2 = c1**2
            safe_aq = np.max([aq, np.ones(len(aq))*1e-6], axis=0)
            sq = 1 + bq*c1/safe_aq
            h_sq = 1 - K*phi(qv, 2*L*R)
            with plt.Dp():
                fig, (ax0, ax1, ax2) = plt.subplots(ncols=3, figsize=(18,6))
                fig.suptitle("estimate_RgL result: $R_g=%.3g$" % Rg, fontsize=20)
                ax1.set_title("Debug Plot on the B(q) basis", fontsize=16)
                ax2.set_title("Debug Plot on the S(q) basis", fontsize=16)

                ax1.plot(qv, bq*c2, label="Data B(q)")
                ax1.plot(qv, h_bq*c2, label="Sphere B(q)")
                ymin, ymax = ax1.get_ylim()
                ax1.set_ylim(ymin, ymax)
                ax1.plot(qv, -bound*c2, ":", color="red")
                ax1.plot(qv, +bound*c2, ":", color="red")
                ax1.set_xlim(0, qv[i2])
                # ax1.set_ylim(-20*c2, 20*c2)
                ax1.legend()

                ax2.plot(qv, sq, label="Data Sq(q)")
                ax2.plot(qv, h_sq, label="Sphere Sq(q)")
                ax2.set_xlim(0, qv[i2])
                ymin, ymax = ax2.get_ylim()
                ax2.set_ylim(-0.05, 5)
                ax2.legend()

                fig.tight_layout()
                plt.show()

        return Rg, R, retL, hK, hL



"""

    ZeroExExcelFormatter.py

    Copyright (c) 2017-2025, SAXS Team, KEK-PF

"""
from molass_legacy.KekLib.ExcelCOM import (
        xlCategory, xlValue, xlLine,
        msoThemeColorText1, msoThemeColorAccent1,
        xlUpward, xlLow,
        msoTrue, xlXYScatter,
        RGB,
        xlMaximum )
from molass_legacy.KekLib.ExcelLogChart import ExcelLogChart
from molass_legacy.KekLib.HtmlColorNames import *
from .DefaultFont import set_default_bookfont_com

ADD_ERROR_BARS = True
ATSAS_COLOR = darkgray

def add_boundary_line( ws, chart, info ):
    j, text, yoffset = info
    if j is None:
        return

    # print( 'add_boundary_line: j=', j, text )

    point = ws.get_chart_data_point( chart, 0, j )
    # print( [j], 'point.Left=', point.Left, ', Point.Width=', point.Width )

    chart_area = chart.ChartArea
    X, Y, H = chart_area.Left, chart_area.Top, chart_area.Height
    x, y = point.Left, point.Top
    plot_area = chart.PlotArea
    pt, ph  = plot_area.Top, plot_area.Height

    # print( 'chart_area.Top=', chart_area.Top, 'chart_area.Height=', chart_area.Height )
    # print( 'plot_area.Top=', plot_area.Top, 'plot_area.Height=', plot_area.Height )

    line_bottom = Y+pt+ph+yoffset
    line = ws.draw_line( ( X+x, Y+pt+30 ), ( X+x, line_bottom ) )
    textbox = ws.draw_textbox( text, ( X+x-50, line_bottom ), 250, 24, visible=False )
    textbox.TextFrame2.TextRange.Font.Size = 16

def add_format_setting_to_extrapolated_chart( ws, chart, logscale=False, boundaries=[] ):
    
    # plot area frame lines
    line = chart.PlotArea.Format.Line
    line.Visible    = True
    line.ForeColor.ObjectThemeColor = msoThemeColorText1

    # grid lines
    chart.Axes(xlValue).MajorGridlines.Format.Line.Visible = False
    chart.Axes(xlCategory).MajorGridlines.Format.Line.Visible = False

    # text font size
    chart.ChartArea.Format.TextFrame2.TextRange.Font.Size = 16

    if logscale:
        y_axis = chart.Axes(xlValue)
        # y_axis.TickLabels.NumberFormatLocal = "#,##0.000_)"
        # y_axis.TickLabels.NumberFormatLocal = "0.0E+00"
        y_axis.MajorUnit    = 10
        chart.Axes(xlCategory).TickLabelPosition = xlLow

        log_chart = ExcelLogChart( chart, ws.worksheet )
        log_chart.change_y_labels( superscripts=True )
        L, T, W, H = log_chart.get_plotarea()
        # print( 'L, T, W, H=', L, T, W, H )
        log_chart.change_plotarea( left=L+30, width=W-50 )
        log_chart.change_yaxis_title_area( left=25 )

    for info in boundaries:
        add_boundary_line( ws, chart, info )

class ZeroExReportArgs:
    def __init__(self, title, book_obj, book_file, axis_direction_desc):
        self.title = title
        self.ud = book_obj.ud
        self.indeces = book_obj.indeces
        self.c_max = book_obj.c_max
        self.overlap_from_max = book_obj.overlap_from_max
        self.guinier_boundary = book_obj.guinier_boundary 
        self.boundary_j = book_obj.boundary_j
        self.lrf_boundary_j = book_obj.lrf_boundary_j
        self.num_rows = book_obj.num_rows
        self.need_bq = book_obj.need_bq
        self.book_file = book_file
        self.axis_direction_desc = axis_direction_desc

def add_result_format_setting(excel_client_, args, logger):
    # logger is not yet used

    book_file = args.book_file
    axis_direction_desc = args.axis_direction_desc

    ws = excel_client_.openWorksheet( book_file + '(' + args.title  + ')' )
    set_default_bookfont_com(ws.workbook)

    colors = [ green, orange, ATSAS_COLOR ]

    for i in range(3):
        p_chart = ws.get_chart( i )

        if i == 0:
            for j in [1, 2]:
                # TODO: move this to openpyxl
                series = p_chart.FullSeriesCollection(j)
                series.MarkerSize = 5
            if args.ud == 1 and axis_direction_desc == 1:
                axis = p_chart.Axes(xlCategory)
                axis.ReversePlotOrder = True
                axis.Crosses = xlMaximum

        # plot area frame lines
        line = p_chart.PlotArea.Format.Line
        line.Visible    = True
        line.ForeColor.ObjectThemeColor = msoThemeColorText1

        # grid lines
        p_chart.Axes(xlValue).MajorGridlines.Format.Line.Visible = False

        # text font size
        p_chart.ChartArea.Format.TextFrame2.TextRange.Font.Size = 12

        category = p_chart.Axes(xlCategory)
        tick_labels = category.TickLabels

        if i == 0:
            max_no = args.indeces[-1]
            if max_no >= 100:
                tick_labels.Orientation     = xlUpward
        else:
            tick_labels.NumberFormatLocal   = "#,##0.000_)"
            tick_labels.Orientation         = xlUpward
            if ADD_ERROR_BARS:
                for j in range(2):
                    series = p_chart.SeriesCollection(j+1)
                    series.ErrorBars.Format.Line.ForeColor.ObjectThemeColor = msoThemeColorAccent1

        # TODO: move this to openpyxl
        if i == 0:
            j_jist = [ 1, 2 ]
        else:
            j_jist = [ 1, 2, 3 ]
            p_chart = ws.get_chart( i )
            p_chart.ChartType = xlXYScatter
            axis = p_chart.Axes(xlCategory)
            axis.MaximumScale = args.c_max * 1.05
            axis.MinimumScale = 0

            if args.ud == 1 and axis_direction_desc == 0:
                axis.ReversePlotOrder = True
                axis.Crosses = xlMaximum

        for j in j_jist:
            series = p_chart.FullSeriesCollection(j)
            line = series.Format.Line
            line.Visible = msoTrue
            line.ForeColor.RGB = RGB( colors[ j-1 ] )   # necessary to remove the border. better way?
            line.Weight = 1
            # series.MarkerStyle = -4142  # no marker
            series.MarkerSize = 5
            p = 1 if args.ud == 0 else series.Points().Count
            point = series.Points(p)
            point.MarkerSize = 10

    atsas_j = args.overlap_from_max
    print( "boundary_j=", args.boundary_j )

    boundaries = []

    if args.lrf_boundary_j is None:
        if atsas_j is None or atsas_j > args.num_rows - 1:
            print( 'WARNING: ATSAS gave an exceeding boundary ', atsas_j )
            # TODO: annotate in the chart
        else:
            boundaries.append( [ atsas_j, "ATSAS Boundary", -160 ] )
    else:
        boundaries.append( [ args.lrf_boundary_j , "Rank Boundary", -80 ] )

    boundaries.append( [ args.guinier_boundary , "Guinier Boundary", -120 ] )
    boundaries.append( [ args.boundary_j , "Regression Boundary", -80 ] )

    ab_chart = ws.get_chart( 3 )
    add_format_setting_to_extrapolated_chart( ws, ab_chart, logscale=True, boundaries=boundaries )

    if args.need_bq:
        ab_chart = ws.get_chart( 4 )
        add_format_setting_to_extrapolated_chart( ws, ab_chart  )

    ws.zoom( 70 )
    scroll_column=16
    ws.freeze_panes( "B2", scroll_column=scroll_column )

    ws.workbook.Save()
    ws.workbook.Close()     # required in the more_multicore mode

class ZeroExOverlayArgs:
    def __init__(self, title, book_obj, book_file, boundary_indeces):
        self.title = title
        self.need_bq = book_obj.need_bq
        self.book_file = book_file
        self.boundary_indeces = boundary_indeces

def add_overlay_format_setting( excel_client, args, logger):
    # logger is not yet used

    need_bq = args.need_bq
    book_file = args.book_file
    boundary_indeces = args.boundary_indeces

    ws = excel_client.openWorksheet( book_file + '(' + args.title  + ')' )
    set_default_bookfont_com(ws.workbook)

    boundaries = []
    yoffset = -120
    for j, b in enumerate( boundary_indeces ):
        ad = 'Asc' if j == 0 else 'Desc'
        boundaries.append( [ b , "Regression Boundary (%s)" % ad, yoffset ] )
        yoffset += 40

    aq_chart = ws.get_chart( 0 )
    add_format_setting_to_extrapolated_chart( ws, aq_chart, logscale=True, boundaries=boundaries )

    if need_bq:
        bq_chart = ws.get_chart( 1 )
        add_format_setting_to_extrapolated_chart( ws, bq_chart  )

    ws.zoom( 70 )
    ws.freeze_panes( "B2", scroll_column=6 )

    ws.workbook.Save()
    ws.workbook.Close()     # required in the more_multicore mode



"""

    1) python create_icon_image.py
    2) convert at https://converticon.com/
    3) as administrator
       rcedit.exe path_dir\serial_analyzer.exe --set-icon path_dir\serial_analyzer.ico

       F:\Tools\rcedit\rcedit-master\Default\rcedit.exe

"""
import numpy                as np
import matplotlib.pyplot    as plt
from scipy                  import misc
from scipy.stats            import norm

image = np.zeros( ( 64, 64, 4 )  )
image[:, :] =  [ 255, 255, 255, 255 ]

x   = np.arange( 64, dtype=int )
y   = norm.pdf( x, 32, 6 )
y_  = 60 * y / y[32]

for i in x:
    j = 62 - int(y_[i])
    js_ = slice( max(0, j-1), min(64, j+2) )
    is_ = slice( max(0, i-1), min(64, i+2) )
    image[ js_, is_ ] = [ 0, 0, 255, 255 ]

image[ 62:64, : ] = [ 255, 0, 0, 255 ]
image[ 0, : ] = [ 127, 127, 127, 255 ]
image[ :, 0 ] = [ 127, 127, 127, 255 ]
image[ :, 63 ] = [ 127, 127, 127, 255 ]

misc.imsave( 'icon.png', image )

plt.imshow( image )
plt.tight_layout()
plt.show()



# coding: utf-8
"""
    PreprocessDialog.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import os
import numpy as np
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from MethodAdjusterFrame import MethodAdjusterFrame
from SubtractorFrame import SubtractorFrame

class PreprocessDialog(Dialog):
    def __init__(self, parent):
        Dialog.__init__(self, parent, "Microfluidic Preprocess", visible=False)

    def body(self, body_frame):
        tk_set_icon_portable(self)

        self.notebook = ttk.Notebook(body_frame)
        self.notebook.pack()

        mtd_frame = MethodAdjusterFrame(body_frame)
        self.notebook.add(mtd_frame, text='Method File Ajuster')

        subt_frame = SubtractorFrame(body_frame)
        self.notebook.add(subt_frame, text='Intensity Subtractor')

        current_tab = self.notebook.tabs()[-1]
        self.notebook.select(current_tab)

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack(fill=Tk.X, expand=1, padx=5)
        bottom_space = Tk.Frame(self, height=10)
        bottom_space.pack()

        w = Tk.Button(box, text="▼ Exit", width=10, command=self.exit)
        w.pack(side=Tk.LEFT, padx=10, pady=5)

        w = Tk.Button(box, text="▶ Main", width=10, command=self.ok)
        w.pack(side=Tk.RIGHT, padx=10, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def show(self):
        self._show(wait=False)
        self.waited_widget = Tk.Frame(self)
        self.wait_window(self.waited_widget)

    def ok(self):
        self.apply()
        self.close()

    def exit(self):
        self.close()

    def close(self):
        self.withdraw()
        self.update_idletasks()
        self.grab_release()
        self.waited_widget.destroy()

    def apply(self):
        self.applied = True



# coding: utf-8
"""
    MctDecompDialog.py

    Copyright (c) 2019-2020, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from lmfit import Parameters, minimize
import matplotlib.pyplot as plt
import matplotlib.patches as mpl_patches
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk
from molass_legacy.KekLib.TkUtils import split_geometry
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable, BlinkingFrame
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry
from molass_legacy.KekLib.TkUtils import is_low_resolution
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting, get_xray_picking
from MicrofluidicElution import get_mtd_elution
from TwoStateSolver import TwoStateSolver
from ThreeStateMonomerSolver import ThreeStateMonomerSolver
from ThreeStateDimerSolver import ThreeStateDimerSolver
from PreviewButtonFrame import PreviewButtonFrame
from SvdDenoise import get_denoised_data

MODEL_ID_LIST = ["One State", "Two-State", "Three-State Monomer", "Three-State Dimer"]
DRAW_COEFF_VECTORS = False

class MctDecompDialog(Dialog):
    def __init__(self, parent, xdata, in_folder):
        self.logger = logging.getLogger(__name__)
        self.busy = False
        self.applied = False
        self.xdata = xdata
        self.mapper = None
        self.in_folder = in_folder
        self.popup_menu = None
        self.popup_label0 = None
        self.popup_label1 = None
        set_setting('conc_factor', 5)   # TODO: change setting according to data date
        Dialog.__init__(self, parent, "Micorfluidic Decomposer", visible=False)

    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()
        tbframe = Tk.Frame(body_frame)
        tbframe.pack(fill=Tk.X, expand=1)
        iframe = Tk.Frame(body_frame)
        iframe.pack(fill=Tk.X, expand=1, padx=20, pady=10)

        self.build_canvas(cframe, tbframe)
        self.builf_param_widgets(iframe)
        self.draw_init()

    def get_canvas_width( self ):
        return int( self.mpl_canvas_widget.cget('width'))

    def build_canvas(self, cframe, tbframe):
        figsize = (17,6) if is_low_resolution() else (23,8)
        fig = plt.figure(figsize=figsize)
        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.mpl_canvas.mpl_connect('button_press_event', self.on_mpl_button_press)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tbframe)
        self.toolbar.update()

        ax1 = fig.add_subplot(131, projection='3d') # it seems this must be done after creation of self.mpl_canvas
        ax2 = fig.add_subplot(132)
        ax3 = fig.add_subplot(133)
        axt = ax3.twinx()
        self.axins = None
        axes = [ax1, ax2, ax3, axt]
        self.fig = fig
        self.axes = axes

        fig.subplots_adjust(top=0.88, bottom=0.1, left=0.03, right=0.95, wspace=0.2)

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack(fill=Tk.X, expand=1, padx=5)
        bottom_space = Tk.Frame(self, height=10)
        bottom_space.pack()

        for j in range(2):
            box.columnconfigure(j, weight=1)

        w = Tk.Button(box, text="◀ Main", command=self.cancel)
        w.grid(row=0, column=0, sticky=Tk.W, padx=10, pady=5)
        self.back_button = w

        w = Tk.Button(box, text="▶ Serial Analysis", command=self.ok)
        w.grid(row=0, column=1, sticky=Tk.E, padx=10, pady=5)
        self.next_button = w

    def show(self):
        self._show(wait=False)

        fully_automatic = get_setting( 'fully_automatic' )
        if fully_automatic == 1:
            self.parent.after(2000, lambda: self.next_button.invoke() )

        self.waited_widget = Tk.Frame(self)
        self.wait_window(self.waited_widget)

    def ok(self):
        self.apply()
        self.close()

    def exit(self):
        self.close()

    def close(self):
        self.withdraw()
        self.update_idletasks()
        self.grab_release()
        self.waited_widget.destroy()

    def apply(self):
        self.preview_frame.update_settings()
        self.applied = True

    def draw_init(self):
        xdata = self.xdata
        in_folder = self.in_folder
        num_files = xdata.data.shape[1]
        mtd_elution = get_mtd_elution(in_folder, num_files)
        self.slice_ = slice_ = mtd_elution.guess_range(num_files)
        ax1, ax2 = self.axes[0:2]
        self.fig.suptitle("Decomposition for " + in_folder, fontsize=20)
        xdata.plot(ax1)
        self.draw_elution(ax2, xdata, mtd_elution, slice_)

    def decompose(self, draw_decomposition=False):
        model_id = self.model_id.get()
        if model_id == MODEL_ID_LIST[0]:
            return

        self.solve_btn_blink.stop()

        self.drawing_curves = False
        self.showing_detail = False
        ax3, axt = self.axes[2:4]
        ax3.cla()
        axt.cla()
        axt.grid(False)
        if self.axins is not None:
            self.axins.remove()
            self.axins = None

        xdata = self.xdata

        self.ax3_title = 'Decomposition using %s Model' % model_id
        ax3.set_title(self.ax3_title, fontsize=16)

        model_id = self.model_id.get()
        if model_id == MODEL_ID_LIST[1]:
            solver = TwoStateSolver()
            data = xdata.data[:,self.slice_]
            solver.solve(data)
            self.solver = solver
            self.draw_decomposition(ax3, data, xdata.e_index, axt)
            if DRAW_COEFF_VECTORS:
                self.draw_coeff_vectors(ax3, self.slice_)
        elif model_id == MODEL_ID_LIST[2]:
            solver = ThreeStateMonomerSolver()
            data = xdata.data[:,self.slice_]
            solver.solve(data)
            self.solver = solver
            self.draw_decomposition(ax3, data, xdata.e_index, axt)
        elif model_id == MODEL_ID_LIST[3]:
            solver = ThreeStateDimerSolver()
            data = xdata.data[:,self.slice_]
            solver.solve(data)
            self.solver = solver
            self.draw_decomposition(ax3, data, xdata.e_index, axt)
        self.mpl_canvas.draw()

    def get_model(self):
        return self.solver.get_fit_model()

    def draw_elution(self, ax, xdata, mtd_elution, slice_):
        pick_pos = get_xray_picking()
        ax.set_title("Elution at Q=%.2g" % pick_pos, fontsize=16)
        e_y = xdata.e_y
        ax.plot(e_y, color='orange')
        axt = ax.twinx()
        axt.grid(False)
        # x, y = mtd_elution.get_elution_data(len(e_y))
        t, x, y = mtd_elution.make_simulation_data()
        axt.plot(x, y, ':', color='blue', alpha=0.3)
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax)

        f, t = [slice_.start, slice_.stop]
        for k, p in enumerate([f, t]):
            ax.plot([p, p], [ymin, ymax], ':', color='gray')
            self.range_vars[k].set(p)

        rect = mpl_patches.Rectangle(
                (f, ymin),      # (x,y)
                t - f,          # width
                ymax - ymin,    # height
                facecolor   = 'cyan',
                alpha       = 0.2,
            )
        ax.add_patch(rect)

    def builf_param_widgets(self, iframe):

        for j in range(3):
            iframe.columnconfigure(j, weight=1)

        width = self.get_canvas_width()//3
        print('width=', width)

        space_frame = Tk.Frame(iframe, width=width)
        space_frame.grid(row=0, column=0)

        range_frame = Tk.Frame(iframe, width=width)
        range_frame.grid(row=0, column=1, sticky=Tk.N)

        decomp_frame = Tk.Frame(iframe, width=width)
        decomp_frame.grid(row=0, column=2, sticky=Tk.N)

        range_label = Tk.Label(range_frame, text="Range Selection: ")
        range_label.grid(row=0, column=0)

        j_max = len(self.xdata.j) - 1

        self.range_vars = []
        for k, t in enumerate(["from", "to"]):
            label = Tk.Label(range_frame, text=t)
            label.grid(row=0, column=1+k*2)
            var = Tk.IntVar()
            self.range_vars.append(var)
            spinbox = Tk.Spinbox( range_frame, textvariable=var,
                                    from_=0, to=j_max, increment=1,
                                    justify=Tk.CENTER, width=6 )
            spinbox.grid(row=0, column=2+k*2, padx=10)

        model_label = Tk.Label(decomp_frame, text="Model Selection: ")
        model_label.grid(row=0, column=0)

        self.model_id = Tk.StringVar()
        self.model_id.set(MODEL_ID_LIST[1])
        model_id_box = ttk.Combobox(decomp_frame, textvariable=self.model_id, width=24, justify=Tk.CENTER)
        model_id_box[ 'values' ] = MODEL_ID_LIST
        model_id_box.grid(row=0, column=1)

        self.model_id.trace('w', self.model_id_tracer)

        self.solve_btn_blink = BlinkingFrame(decomp_frame)
        self.solve_btn_blink.grid(row=0, column=2)
        self.solve_btn = Tk.Button(self.solve_btn_blink, text="Solve", command=self.decompose)
        self.solve_btn.pack()
        self.solve_btn_blink.objects = [self.solve_btn]

        space = Tk.Frame(decomp_frame, height=20)
        space.grid(row=1, column=0)

        self.preview_frame = PreviewButtonFrame(decomp_frame, bd=3, relief=Tk.RIDGE, dialog=self)
        self.preview_frame.grid(row=2, column=0, columnspan=3)

    def model_id_tracer(self, *args):
        model_id = self.model_id.get()
        print(model_id)
        self.solve_btn_blink.start()
        if model_id == MODEL_ID_LIST[2]:
            self.preview_frame.set_guard_for_three_state_model()
        else:
            self.preview_frame.remove_guard_for_three_state_model()

    def on_mpl_button_press(self, event):
        if event.xdata is None:
            return

        if event.button == 3:
            self.update_popup_menu()
            self.create_popup_menu()
            w, h, x, y = split_geometry(self.geometry())
            self.popup_menu.post(x + event.x + 20, y + h - event.y - 180)
            # TODO: do this without dialog dependent params 20, 180
            return

    def create_popup_menu(self):
        if self.popup_menu is None:
            self.popup_menu = Tk.Menu( self, tearoff=0 )
            self.popup_menu.add_command(label=self.popup_label0, command=self.redraw)
            self.popup_menu.add_command(label=self.popup_label1, command=lambda: self.redraw(toggle=False, toggle_detail=True))
            self.popup_menu.add_command(label="Inspection in various angles", command=self.inspection_dialog)
        else:
            self.popup_menu.entryconfigure(0, label=self.popup_label0)
            self.popup_menu.entryconfigure(1, label=self.popup_label1)
        state1 = Tk.NORMAL if not self.drawing_curves and self.model_id.get() == MODEL_ID_LIST[1] else Tk.DISABLED
        self.popup_menu.entryconfigure(1, state=state1)

    def update_popup_menu(self):
        self.popup_label0 = 'Draw with bands' if self.drawing_curves else 'Draw with curves'
        self.popup_label1 = ('Hide' if self.showing_detail else 'Show') + ' denaturant dependency'

    def redraw(self, toggle=True, toggle_detail=False):
        xdata = self.xdata
        data = xdata.data[:,self.slice_]

        ax3, axt = self.axes[2:4]
        for ax in [ax3, axt]:
            ax.cla()
        ax3.set_title(self.ax3_title, fontsize=16)
        axt.grid(False)
        if toggle:
            self.drawing_curves ^= True
        if toggle_detail:
            self.showing_detail ^= True
        self.draw_decomposition(ax3, data, xdata.e_index, axt)

    def draw_decomposition(self, ax, data, e_index, axt, toggle_detail=False):
        if self.drawing_curves:
            self.solver.plot_components(ax, data, e_index, axt=axt, start=self.slice_.start)
        else:
            self.solver.draw_components_with_bands(ax, data, e_index, start=self.slice_.start, detail=self.showing_detail)
        self.mpl_canvas.draw()
        self.drawn_info = (self.ax3_title, data, e_index, self.slice_.start, self.showing_detail)

    def draw_coeff_vectors(self, ax, slice_):
        e_index = self.xdata.e_index
        values = self.solver.P[e_index,:]
        xspan = max(abs(values[0]), abs(values[2]))*2
        yspan = max(abs(values[1]), abs(values[3]))*2

        side = self.solver.guess_blank_location(slice_)

        if side == 'left':
            bbox = (0.1, 0.2, 1, 1)
            loc = 'lower left'
        else:
            bbox = (0, 0.2, 0.9, 1)
            loc = 'lower right'

        axins = inset_axes(ax, width="30%", height="30%",
                            bbox_to_anchor=bbox,
                            bbox_transform=ax.transAxes, loc=loc)
        axins.set_xticklabels([])
        axins.set_yticklabels([])
        axins.set_facecolor('xkcd:cream')
        axins.patch.set_alpha(0.3)
        axins.grid(False)
        axins.set_title("Coefficient Vectors", fontsize=14)
        axins.plot(0, 0, 'o', color='black', markersize=3)
        axins.set_xlim(-xspan, xspan)
        axins.set_ylim(-yspan, yspan)
        for t, p in [['(af,bf)', values[0:2]], ['(au,bu)', values[2:4]]]:
            x, y = p
            print(t, x, y)
            axins.text(x, y, t)
            axins.annotate('', xy=(x, y), xytext=(0,0),
                arrowprops=dict(headwidth=5, headlength=6, width=0.5, color='black', shrink=0)
               )
        self.axins = axins

    def get_decomp_info(self):
        from MctRangeInfo import MctRangeInfo
        conc = self.get_conc()
        ranges = self.make_alt_ranges()
        curves = self.get_conc_curves()
        return MctRangeInfo(conc, ranges, curves)

    def get_conc(self):
        # return scaled C list

        model_id = self.model_id.get()

        C = self.solver.C
        y = self.xdata.e_curve.y[self.slice_]

        if model_id == MODEL_ID_LIST[1]:
            scale1 = y[0]/C[0,0]
            scale2 = y[-1]/C[2,-1]
            rest_list = [scale1*C[0,:], scale1*C[1,:], scale2*C[2,:], scale2*C[3,:]]
        elif model_id == MODEL_ID_LIST[2]:
            scale1 = y[0]/C[0,0]
            m = len(y)//2
            scale2 = y[m]/C[1,m]
            scale3 = y[-1]/C[2,-1]
            rest_list = [scale1*C[0,:], scale2*C[1,:], scale3*C[2,:]]
        else:
            assert False

        return rest_list

    def get_conc_curves(self):
        return self.solver.get_modeled_curves(self.xdata.e_index)

    def make_range_info(self):
        return None

    def make_alt_ranges(self):
        from molass_legacy.DataStructure.PeakInfo import PeakInfo
        from molass_legacy.DataStructure.AnalysisRangeInfo import PairedRange

        start, stop = self.slice_.start, self.slice_.stop

        if False:
            import molass_legacy.KekLib.DebugPlot as dplt
            from molass_legacy.Elution.CurveUtils import simple_plot

            e_curve = self.xdata.e_curve
            fig = dplt.figure()
            ax = fig.gca()
            ax.set_title("make_range_info debug")
            simple_plot(ax, e_curve)

            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)
            for x in [start, stop]:
                ax.plot([x,x], [ymin, ymax], ':', color='gray')

            dplt.show()

        """
        make a range analogous to a SEC peak range
        """
        ret_list = []

        pno = 0
        pinfo = PeakInfo(pno, start)
        range_list = [[start, stop-1]]
        ret_list.append(PairedRange(pinfo, *range_list))

        model_id = self.model_id.get()
        if model_id == MODEL_ID_LIST[2]:
            itx = int(start + self.solver.get_imermediate_topx())
            pno += 1
            pinfo = PeakInfo(pno, itx)
            range_list = [[start, stop-1]]
            ret_list.append(PairedRange(pinfo, *range_list))

        pno += 1
        pinfo = PeakInfo(pno, stop-1)
        range_list = [[start, stop-1]]
        ret_list.append(PairedRange(pinfo, *range_list))

        return ret_list

    def inspection_dialog(self):
        from InspectionInAngles import InspectionInAnglesDialog
        dialog = InspectionInAnglesDialog(self, self.xdata.vector, self.xdata.e_index, self.drawn_info)
        dialog.show()



# coding: utf-8
"""
    ElutionMixtureModel.py

    learned at
    Gaussian Mixture Models of an Image's Histogram
    https://stackoverflow.com/questions/45805316/gaussian-mixture-models-of-an-images-histogram

    Copyright (c) 2019-2020, SAXS Team, KEK-PF
"""
import numpy as np
from sklearn.mixture import GaussianMixture
try:
    # for numba 1.49 or later
    from numba.core.decorators import jit
except:
    from numba.decorators import jit
from lmfit import minimize, Parameters
import molass_legacy.KekLib.DebugPlot as plt

@jit(nopython=True)
def generate_samples(x, y, n):
    data = []
    for k, i in enumerate(x):
        for j in range(int(y[k]*n)):
            data.append(i)

    return np.array(data)

def gaussian(x, mu, sigma, scale):
    return scale * np.exp( - (x - mu)**2 / (2 * sigma**2) )

class ElutionMixtureModel:
    def __init__(self, n_components=3):
        self.gmm = GaussianMixture(n_components, max_iter=1000, covariance_type='spherical' )

    def fit_elution(self, curve, slice_=None, precision=3):
        self.curve = curve
        x_ = curve.x
        y_ = curve.y / curve.max_y

        if slice_ is None:
            x = x_
            y = y_
        else:
            x = x_[slice_]
            y = y_[slice_]

        n = 10**precision
        self.data = generate_samples(x, y, n)
        self.gmm.fit(X=np.expand_dims(self.data,1))

        # Evaluate GMM
        # gmm_x = np.linspace(0,253,256)
        self.gmm_x = x
        self.gmm_y = np.exp(self.gmm.score_samples(self.gmm_x.reshape(-1,1)))
        self.y_ = y_

    def get_gaussians(self):
        x = self.gmm_x
        y = self.gmm_y

        gy_list = []
        ty = np.zeros(len(x))
        for k in range(len(self.gmm.weights_)):
            w = self.gmm.weights_[k]
            m = self.gmm.means_[k][0]
            v = self.gmm.covariances_[k]
            print([k], w, m, v)
            gy = gaussian(x, m, np.sqrt(v), w)
            gy_list.append(gy)
            ty += gy

        def obj_func(params):
            S   = params['S']
            return ty*S - y

        params = Parameters()
        S_init = np.max(y)/np.max(ty)
        params.add('S', value=S_init, min=0, max=S_init*100 )
        result = minimize(obj_func, params, args=())

        scale = result.params['S'].value
        return [ scale*gy for gy in gy_list ]

    def plot(self, ax):
        x = self.gmm_x
        y = self.gmm_y
        # Plot histograms and gaussian curves
        r = [x[0], x[-1]+1]
        ny_ = self.y_ / np.sum(self.y_)
        ax.plot(x, ny_, color="orange", label="data")
        ax.hist(self.data, x[-1], r, density=True, alpha=0.5)
        ax.plot(x, y, ':', color="red", lw=3, label="GMM")
        ax.set_ylabel("Intensity")
        ax.set_xlabel("Elution No")

        for k, gy in enumerate(self.get_gaussians()):
            ax.plot(x, gy, label='%d' % k)

        ax.legend()

def gmm_demo(curve):
    num_comp_matrix = np.array([[1, 2, 3], [4, 5, 6]])
    nrows, ncols = num_comp_matrix.shape

    fig, axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=(18, 10))

    for i in range(nrows):
        for j in range(ncols):
            n = num_comp_matrix[i, j]
            ax = axes[i, j]
            ax.set_title("GMM with n_components=%d" % n)
            gmm = ElutionMixtureModel(n_components=n)
            # slice_=slice(150, 400)
            slice_=None
            gmm.fit_elution(curve, slice_=slice_, precision=3)
            gmm.plot(ax)

    fig.tight_layout()
    plt.show()



"""
    Optimizer.ParameterTransitionPlot.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.DebugPlot import push, pop, get_parent
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
from molass_legacy.KekLib.ScrolledFrame import ScrolledFrame
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Optimizer.FvScoreConverter import convert_scores
from molass_legacy.Optimizer.ParamsIterator import create_iterator

class ParameterTransitionPlot(Dialog):
    def __init__(self, parent, optimizer, params_array, bounds_array, fv_list):
        self.optimizer = optimizer
        self.params_array = params_array
        self.bounds_array = bounds_array
        self.fv_list = fv_list
        Dialog.__init__(self, parent, "Parameter Transition Plot", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        title_frame = Tk.Frame(body_frame)
        title_frame.pack(side=Tk.TOP, fill=Tk.X, expand=1)
        in_folder = get_in_folder()
        result_folder = get_setting('optimizer_folder').replace('/optimized', '')
        func_name = self.optimizer.get_name()
        title = Tk.Label(title_frame,
                         text="Parameter Transition on %s from %s with %s" % (in_folder, result_folder, func_name),
                         bg="white", font=(None,20))
        title.pack(fill=Tk.X, expand=1)
        sv_frame = ScrolledFrame(body_frame)
        sv_frame.pack(side=Tk.TOP)

        self.btn_frame = Tk.Frame(body_frame)
        self.btn_frame.pack(side=Tk.BOTTOM)
        tool_frame = Tk.Frame(body_frame)
        tool_frame.pack(side=Tk.BOTTOM)
        pv_frame = ScrolledFrame(body_frame)
        pv_frame.pack(side=Tk.BOTTOM)

        sv_frame.set_sync_info(pv_frame.canvas_info)
        pv_frame.set_sync_info(sv_frame.canvas_info)

        params_array = self.params_array
        nrows, ncols = params_array.shape

        width = 40
        sv_fig, sv_ax = plt.subplots(figsize=(width, 3))
        pv_fig, pv_axes = plt.subplots(nrows=ncols, figsize=(width,ncols*3))
        self.draw_transition(sv_fig, sv_ax, pv_fig, pv_axes)
        self.sv_canvas = FigureCanvasTkAgg(sv_fig, sv_frame.interior)
        self.sv_canvas_widget = self.sv_canvas.get_tk_widget()
        self.sv_canvas_widget.pack(fill=Tk.X, expand=1)
        self.pv_canvas = FigureCanvasTkAgg(pv_fig, pv_frame.interior)
        self.pv_canvas_widget = self.pv_canvas.get_tk_widget()
        self.pv_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.pv_canvas, tool_frame)
        self.toolbar.update()
        self.after(500, lambda: self.geometry("2000x800"))

    def buttonbox(self):
        box = self.btn_frame
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

    def draw_transition(self, sv_fig, sv_ax, pv_fig, pv_axes):
        sv_list = []
        minima_points_list = []
        start = 0
        for fv_array in self.fv_list:
            sv_list.append(convert_scores(fv_array))
            for i in create_iterator(fv_array):
                minima_points_list.append(start + i)
            start += len(fv_array)
        niter_vector = np.cumsum([len(v) for v in sv_list])
        minima_points = np.array(minima_points_list)
        def plot_job_sections_with_minima(ax):
            for niter in niter_vector[:-1]:
                ax.axvline(niter, color="gray", alpha=0.3)
            for i in minima_points:
                ax.axvline(i, color="red", ls=':')
        sv_ax.set_ylabel("SV", fontsize=16)
        sv_vector = np.concatenate(sv_list)
        sv_ax.plot(sv_vector)
        plot_job_sections_with_minima(sv_ax)
        sv_ax.set_ylim(0, 100)
        names = self.optimizer.get_parameter_names()
        for ax, name, params in zip(pv_axes, names, self.params_array.T):
            ax.set_ylabel(name, fontsize=16)
            ax.plot(params)
            plot_job_sections_with_minima(ax)

        for fig in [sv_fig, pv_fig]:
            fig.tight_layout()
            fig.subplots_adjust(left=0.03, right=0.97)

def plot_transition_impl(optimizer, params_array, bounds_array, fv_list):
    push()
    parent = get_parent()
    dialog = ParameterTransitionPlot(parent, optimizer, params_array, bounds_array, fv_list)
    dialog.show()
    pop()


"""
    Models.Stochastic.PairedAnalysisDemo.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
from importlib import reload

def prepare_data(input_folders, exec_spec):
    import Batch.BatchUtils
    reload(Batch.BatchUtils)
    from molass_legacy.Batch.BatchUtils import load_lrfsrc

    use_mapping = exec_spec.get('use_mapping')

    info_list = []
    for k, folder in enumerate(input_folders):
        if k == 1 and use_mapping is not None and use_mapping:
            from molass_legacy.Batch.BatchMappingUtils import load_lrfsrc_with_mapping
            lrfsrc = load_lrfsrc_with_mapping(folder, lrfsrc)
        else:
            lrfsrc = load_lrfsrc(folder)
        info_list.append((folder, lrfsrc))
    return info_list

def debug_main():
    import molass_legacy.KekLib.DebugPlot as plt
    import Models.Stochastic.PairedAnalysisSpecs as specs
    reload(specs)
    from molass_legacy.Models.Stochastic.PairedAnalysisSpecs import get_exec_spec

    exec_spec = get_exec_spec(3)
    input_folders   = exec_spec["input_folders"]

    info_list = prepare_data(input_folders, exec_spec)
    def paired_analysis():
        import Models.Stochastic.PairedAnalysis
        reload(Models.Stochastic.PairedAnalysis)
        from molass_legacy.Models.Stochastic.PairedAnalysis import paired_analysis_main
        try:
            paired_analysis_main(info_list, exec_spec)
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(None, "paired_analysis_impl: ")

    extra_button_specs = [
                    ("Paired Analysis", paired_analysis)
                    ]

    with plt.Dp(button_spec=["OK", "Cancel"],
                extra_button_specs=extra_button_specs):
        fig, ax = plt.subplots()
        plt.show()

def demo():
    from molass_legacy.KekLib.OurTkinter import Tk
    root = Tk.Tk()
    root.withdraw()
    def demo_impl():
        debug_main()
        root.quit()
    root.after(100, demo_impl)
    root.mainloop()

if __name__ == "__main__":
    import sys
    sys.path.append("../lib")
    from molass_legacy.KekLib.OurMatplotlib import mpl_1_5_backward_compatible_init
    mpl_1_5_backward_compatible_init()
    demo()


"""

    ScatteringBaseUtil.py

    scattering baseline solver

    Copyright (c) 2017-2022, SAXS Team, KEK-PF

"""
import time
from ScatteringBaseCorrector import ScatteringBaseCorrector, SMALL_ANGLE_LIMIT
from molass_legacy._MOLASS.SerialSettings import get_setting

def apply_baseline_correction_impl( jvector, qvector, intensity_array,
                                    mapped_info=None,
                                    basic_lpm=False,    # 
                                    index=None,
                                    ecurve=None,
                                    progress_cb=None,
                                    return_base=False,
                                    parent=None, with_demo=False, logger=None,
                                    debug_obj=None,
                                    debug=True ):
    if logger is not None:
        t0 = time.time()

    opt_params      = mapped_info.opt_params
    affine_info     = mapped_info.affine_info
    baseline_opt    = opt_params['xray_baseline_opt']
    baseline_type   = opt_params['xray_baseline_type']

    awmf1 = opt_params['xray_baseline_with_bpa']        # seems to be not effective. why?
    awmf2 = get_setting('xray_baseline_with_bpa')
    xray_baseline_with_bpa = awmf1 or awmf2

    if baseline_type == 3:
        # for MF
        baseline_opt = 0    # no LPM correction

    need_adjustment = opt_params['xray_baseline_adjust'] == 1

    corrector = ScatteringBaseCorrector( jvector, qvector, intensity_array,
                                    curve=ecurve,
                                    affine_info=affine_info,
                                    inty_curve_y=ecurve.y,
                                    baseline_opt=baseline_opt,
                                    baseline_type=baseline_type,
                                    need_adjustment=need_adjustment,
                                    parent=parent, with_demo=with_demo )
    xray_baseline_type = get_setting('xray_baseline_type')
    if xray_baseline_type == 0:
        logger.info('no baseline correction.')
        return None

    logger.info("baseline correction with basic_lpm=%s and xray_baseline_type=%d", str(basic_lpm), xray_baseline_type)
    if basic_lpm or xray_baseline_type == 1:
        base = corrector.correct_all_q_planes( progress_cb=progress_cb, return_base=return_base, debug_obj=debug_obj )
    else:
        base = corrector.correct_with_matrix_base( mapped_info, progress_cb=progress_cb, return_base=return_base )

    if xray_baseline_with_bpa and False:
        data = intensity_array[:,:,1].T
        if debug:
            from importlib import reload
            import Baseline.LambertBeerTester
            reload(Baseline.LambertBeerTester)
            from molass_legacy.Baseline.LambertBeerTester import test_compute_base_plane
            BP = test_compute_base_plane(data, index, ecurve, denoise=False)
        else:
            from molass_legacy.Baseline.LambertBeer import compute_base_plane
            BP = compute_base_plane(data, index, ecurve, denoise=False, debug=debug)
        base = BP
        intensity_array[:,:,1] -= BP.T

    if logger is not None:
        t = time.time() - t0
        with_bpa = ' with BPA' if xray_baseline_with_bpa else ''
        logger.info('baseline correction%s took %.3g seconds.', with_bpa, t)

    if return_base:
        return base



"""
    PnoScdMap.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from molass_legacy._MOLASS.SerialSettings import get_setting

class PnoScdMap:
    def __init__(self, sd, cnv_ranges, debug=False):
        """
        using the paired peak_top_x here because
        self.sd.xray_curve.peak_top_x may have been changed
        from this peak_top_x as in 20200121_2

        also note that these coordinates are in relative values,
        i.e., x - x[0]
        """
        self.logger = logging.getLogger(__name__)
        peak_top_x, scd_colors = get_setting('mapper_cd_color_info')
        pno_map = []
        ecurve = sd.get_xray_curve()
        main_y = None
        for pno, prange in enumerate(cnv_ranges):
            x = prange.top_x
            dist = np.abs(peak_top_x - x)
            k = np.argmin(dist)
            if dist[k] < 5:
                # i.e., this pno-th prange corresponds to the k-th point in peak_top_x
                pass
            else:
                # i.e., there is no corresponding point in peak_top_x with this prange
                k = None
            pno_map.append(k)

            y = ecurve.spline(x)
            if main_y is None or y > main_y:
                main_y = y
                self.main_pno = pno

        self.pno_map = pno_map
        self.scd_colors = scd_colors
        self.logger.info("pno_map=%s, scd_colors=%s", str(pno_map), str(scd_colors))
        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            ecurve = sd.xray_curve
            x = ecurve.x
            y = ecurve.y
            print("cnv_ranges=", cnv_ranges)
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.plot(x, y)
                for prange in cnv_ranges:
                    pass
                fig.tight_layout()
                plt.show()

    def get_color(self, pno):
        i = self.pno_map[pno]
        if i is None:
            return '???'
        else:
            return self.scd_colors[i]

    def get_main_peak_rank(self):
        color = self.get_color(self.main_pno)
        return 2 if color in ["red", "yellow", "orange"] else 1



"""

    MapperConstructor.py

    Copyright (c) 2018-2024, SAXS Team, KEK-PF

"""
import copy
import numpy as np
import logging
from molass_legacy.KekLib.OurTkinter             import Tk
import OurMessageBox        as MessageBox
from .MappingParams import get_mapper_opt_params, get_mapper_simplest_params, set_mapper_opt_params
from molass_legacy.KekLib.ExceptionTracebacker   import ExceptionTracebacker, log_exception
from molass_legacy.KekLib.BasicUtils             import Struct
from molass_legacy._MOLASS.SerialSettings         import get_setting, set_setting, INTEGRAL_BASELINE
from molass_legacy.Test.TesterLogger import write_to_tester_log
from .ElutionMapper import ElutionMapper
from molass_legacy.SerialAnalyzer.ElutionCurve import SEDIMENTATION_LIMIT
from molass_legacy.Baseline.LinearAdjustment import LaJudge

DEBUG = True
ACCEPTABLE_ADJ_DEVIATION    = 0.2   # > 0.1 for Gactin
ADJ_DEVIATION_SCALE         = 10
RE_OPTIMIZE_MIN_SCI_LIMIT   = 50

class MockAnalyzerDialog(Tk.Frame):
    def __init__( self, parent ):
        self.serial_data    = Struct( helper_info=None )
        Tk.Frame.__init__( self, parent )
        self.logger         = logging.getLogger( __name__ )

def create_mapper(parent,
                    serial_data,        # analysis_copy or trimmed sd
                    input_sd,           # sd_orig
                    pre_recog,          # created from sd_orig
                    callbacks=None,
                    analyzer_dialog=None, mapper_hook=None,
                    opt_params=None,
                    return_dialog=False, logger=None ):
    """
    moved this process out of ElutionMapperCanvas.__init__
    because the construction of mapper must be complete
    before ElutionMapperCanvas.__init__
    """

    if analyzer_dialog is None:
        analyzer_dialog = MockAnalyzerDialog( parent )

    mapper = ElutionMapper(serial_data, input_sd, pre_recog, callbacks=callbacks)

    # set them simlpy when used_mapping_params exist
    used_mapping_params = get_setting('used_mapping_params')
    if used_mapping_params is not None:
        if logger is not None:
            logger.info("restoring %s from used_mapping_params.", str(used_mapping_params))
        mapper.set_opt_params(used_mapping_params)
        return mapper

    if serial_data.mtd_elution is None:
        pass
    else:
        # 
        absorbance  = mapper.absorbance
        serial_data.mtd_elution.set_baseplane_params_as_a_temp_fix(absorbance)
        mapper.optimize()
        return mapper

    if mapper_hook is None:
        helper_info = serial_data.helper_info
        apply_patch = False
    else:
        helper_info = mapper_hook.helper_info
        apply_patch = mapper_hook.apply_patch

    if opt_params is None:
        opt_params = get_mapper_opt_params()

    ret = optimize_mapper(mapper, opt_params, parent, serial_data, analyzer_dialog, helper_info, apply_patch, logger)
    if ret is None:
        return None

    nrmsd = ret[1]
    adj_dev = mapper.get_adjustment_deviation()
    alternative = False
    if adj_dev > ACCEPTABLE_ADJ_DEVIATION:
        if logger is not None:
            logger.warning('retrying mapper construction with the simplest params becauce of adj_dev=%g' % adj_dev)
        simplest_params = get_mapper_simplest_params()
        alt_mapper  = ElutionMapper(serial_data, input_sd, pre_recog, callbacks=callbacks)
        alt_ret = optimize_mapper(alt_mapper, simplest_params, parent, serial_data, analyzer_dialog, helper_info, apply_patch, logger)
        if alt_ret is None:
            return None

        alt_adj_dev = alt_mapper.get_adjustment_deviation()
        if alt_adj_dev < adj_dev:
            alt_nrmsd = alt_ret[1]
            if alt_nrmsd < nrmsd:
                alternative = True
                mapper = alt_mapper
                adj_dev = alt_adj_dev
                nrmsd = alt_nrmsd
                if logger is not None:
                    logger.warning('adopted mapper with the simplest params')

    judge = LaJudge(mapper)
    judge.modify_opt_params()

    set_mapper_opt_params(mapper.opt_params)

    write_to_tester_log( 'mapped info: adj_dev=%g, nrmsd=%g, alternative=%s\n' % (adj_dev, nrmsd, str(alternative)) )

    serial_data.helper_info = helper_info
    if return_dialog:
        return mapper, analyzer_dialog
    else:
        return mapper

def optimize_mapper(mapper, opt_params, parent, serial_data, analyzer_dialog, helper_info, apply_patch, logger, debug=False):
    # TODO: remove analyzer_dialog

    pre_recog   = serial_data.pre_recog
    absorbance  = mapper.absorbance
    mapping_ok  = False
    auto_make   = True
    if logger is None:
        logger = analyzer_dialog.logger

    xray_baseline_type = get_setting('xray_baseline_type')
    if xray_baseline_type == INTEGRAL_BASELINE:
        opt_baseline_types = [INTEGRAL_BASELINE]
    else:
        uv_baseline_type = get_setting("uv_baseline_type")
        if serial_data.absorbance.shifted_baseline_ok() or uv_baseline_type == 4:
            opt_baseline_types = [0, 1, 4]
        else:
            # e.g., 20200630_6
            opt_baseline_types = [0, 1]

    temp_params = copy.deepcopy(opt_params)

    while not mapping_ok:
        try:
            min_dev_score = None
            selected_type  = None

            # temporarily allow bad_std_diff to cope with bad data such as Factin
            bad_std_diff_ok_save = get_setting( 'bad_std_diff_ok' )
            set_setting( 'bad_std_diff_ok', 1 )

            if debug:
                dev_score_list = []

            for baseline_type in opt_baseline_types:
                absorbance.compute_base_curve(pre_recog, baseline_type)
                temp_params['uv_baseline_opt'] = 0 if baseline_type == 0 else 1
                temp_params['uv_baseline_type'] = baseline_type
                try:
                    mapper.optimize( helper_info=helper_info, apply_patch=apply_patch, opt_params=temp_params )
                    adj_dev = mapper.get_adjustment_deviation()
                    if logger is not None:
                        method_label = 'LPM' if baseline_type == 0 else 'LB'
                        logger.info('optimize_mapper: nrmsd=%.4g, adj_dev=%.4g with baseline_type=%s for absorbance.compute_base_curve' % (mapper.std_diff, adj_dev, method_label) )
                    dev_score = mapper.std_diff + adj_dev * ADJ_DEVIATION_SCALE
                    if debug:
                        dev_score_list.append(dev_score)
                    if min_dev_score is None or dev_score < min_dev_score:
                        min_dev_score = dev_score
                        selected_type = baseline_type
                except Exception as exc:
                    # baseline_type LB fails for 3/7
                    if DEBUG:
                        log_exception(logger, "failed in mapper.optimize: ", n=8)

            if debug:
                import molass_legacy.KekLib.DebugPlot as plt
                with plt.Dp():
                    fig, ax = plt.subplots()
                    ax.set_title("optimize_mapper debug")
                    ax.plot(dev_score_list, "-o")
                    plt.show()

            if min_dev_score is None:
                raise RuntimeError( 'Failed to optimize for both baseline methods for the absorbance data', [ 5 ] )

            if selected_type < opt_baseline_types[-1]:
                # TODO: improve by avoiding this redundant repitition
                logger.warning( 'undesirable repitition of UV-basecurve computation' )
                absorbance.compute_base_curve(pre_recog, selected_type)
                temp_params['uv_baseline_opt'] = 0 if selected_type == 0 else 1
                temp_params['uv_baseline_type'] = selected_type
                mapper.optimize( helper_info=helper_info, apply_patch=apply_patch, opt_params=temp_params )

            set_setting( 'uv_baseline_type', selected_type )
            mapper.opt_params['uv_baseline_type'] = selected_type       # TODO: investigate why this is required

            set_setting( 'bad_std_diff_ok', bad_std_diff_ok_save )

            if logger is not None:
                logger.info("optimize_mapper: %s has been selected for uv_baseline_method with min_dev_score=%.4g" % ( 'LPM' if selected_type == 0 else 'LB', min_dev_score ))

            mapping_ok  = True
        except RuntimeError as exc:
            etb= ExceptionTracebacker()
            print( etb )
            logger.warning('exception in optimize_mapper etb=%s', str(etb))
            auto_make = False       # don't use MappingAutoHelper further to avoid infinite loop
            if helper_info is None:
                MessageBox.showwarning( "Suggestion",
                        "Restricting elution ranges might help.",
                        parent=parent )
                return None
        except:
            etb = ExceptionTracebacker()
            print(etb)
            logger.error( str(etb) )
            MessageBox.showerror( "Error",
                    "Failed to conctruct a mapper.\n"
                    + "Please report this error to the developer.\n"
                    + "Restricting elution ranges mihgt help.\n\n"
                    + str(etb),
                    parent=parent )
            return None

    if mapper.feature_mapped:
        min_sci = np.min(mapper.get_sci_list())
        if min_sci < RE_OPTIMIZE_MIN_SCI_LIMIT:
            sync_options = 0
            mapper.optimize(opt_params=mapper.opt_params, sync_options=sync_options)
            set_setting('mapper_sync_options', sync_options)
            logger.info('re-optimized mapper with "rmsd finally" due to a bad minimun sci: %.2g.' % min_sci)

    return selected_type, min_dev_score



"""
    PreliminaryRecognition.py

    Copyright (c) 2018-2024, SAXS Team, KEK-PF
"""
import copy
import numpy as np
import logging
from molass_legacy.UV.Absorbance import Absorbance
from . import FlowChange, TrimmingInfo, AutoRestrictor
from molass_legacy.UV.PlainCurve import make_secondary_e_curve_at
from molass_legacy.Mapping.CurveSimilarity import CurveSimilarity
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
from molass_legacy.Baseline.BaselineGuinier import change_baseline_correction_defaults
# from molass_legacy.Trimming.PreliminaryRg import get_default_angle_range_impl
from .PreliminaryRg import get_flange_limit, PreliminaryRg, get_small_anlge_limit
from .EndingAnomaly import check_ending_anomaly
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting

SECONDARY_WAVE_LENGTH   = 400
ACCEPTABLE_SIMILARITY_SCORE = 0.8   # > 0.75 for 20170304

def get_slices_impl(listname):
    print("get_slices_impl: listname=", listname)
    trimlist = get_setting(listname)
    if trimlist is None:
        ret_slices = [slice(None, None), slice(None, None)]
    else:
        ret_slices = []
        for triminfo in trimlist:
            ret_slices.append(slice(triminfo.start, triminfo.stop))
        ret_slices.reverse()
    print("get_slices_impl: ret_slices=", ret_slices)
    return ret_slices

class PreliminaryRecognition:
    def __init__(self, sd, recog_only=False, debug=False):
        """
        "Preliminary Recognition" consists of the following parts.

        self.set_info() --------------- (1) UV.PlainCurve.make_secondary_e_curve_at()
                                                Trimming.Sigmoid.guess_bent_sigmoid()
                                                UV.PlainCurve.check_diehardness()

                                        (2) self.flowchange = FlowChange()
                                                .PeakRegion
                                                    Mapping.CurveSimilarity
                                                        Mapping.PeakMapper

                                        (3) 

        self.set_restrict_info() ------ (4) self.ar = AutoRestrictor()

        """
        self.logger = logging.getLogger( __name__ )
        try:
            # for backward compatibility
            self.pre_recog_copy = sd.get_copy(pre_recog=self)
            self.set_info(self.pre_recog_copy, debug=debug)
        except:
            self.pre_recog_copy = None
            self.set_info(sd, debug=debug)
        self.Rg = None
        if recog_only:
            # for testing AutoRestrictor. See test_1117_TrimmingResult.py
            from importlib import reload
            import Trimming
            reload(Trimming)
            from molass_legacy.Trimming import AutoRestrictor
            pass
        else:
            if self.pre_recog_copy is None:
                self.set_restrict_info(sd)
            else:
                self.set_restrict_info(self.pre_recog_copy)

    def set_info(self, sd, debug=False):
        self.q = sd.intensity_array[0,:,0]
        x_curve = sd.get_xr_curve()

        # construction of absorbance (if needed) must be delayed until exclusion is complete
        self.absorbance = absorbance = sd.absorbance

        if not hasattr(sd, 'mtd_elution') or sd.mtd_elution is None:
            data = absorbance.data
            wl_vector = absorbance.wl_vector
            a_curve = absorbance.a_curve
            a_curve2 = make_secondary_e_curve_at(data, wl_vector, a_curve, self.logger)
            try:
                self.flowchange = FlowChange(a_curve, a_curve2, x_curve, debug=debug)
            except:
                self.logger.warning("FlowChange failed. Using NullFlowChange instead.")
                import molass.FlowChange.NullFlowChange
                from importlib import reload
                reload(molass.FlowChange.NullFlowChange)
                from molass.FlowChange.NullFlowChange import NullFlowChange
                self.flowchange = NullFlowChange(a_curve, a_curve2, x_curve) 
            self.cs = self.verify_approx_mapping(self.flowchange.get_similarity())
            check_ending_anomaly(x_curve, a_curve, a_curve2)
        else:
            self.flowchange = sd.mtd_elution.create_flow_change_proxy()
            self.cs = self.flowchange.similarity

        self.mapped_info = self.cs.mapped_info
        self.get_default_angle_range(sd)

    def verify_approx_mapping(self, similarity):
        # remake if it is bad
        score = similarity.compute_whole_similarity()
        print('score=', score)
        fc_list = self.flowchange.get_real_flow_changes()
        if score < ACCEPTABLE_SIMILARITY_SCORE:
            self.logger.info("re-constructing CurveSimilarity due to poor score %.2g and flowchages %s" % (score, str(fc_list)) )
            a_vector = copy.deepcopy(similarity.a_curve.y)
            for k, fc in enumerate(fc_list):
                if fc is None:
                    continue
                if k == 0:
                    a_vector[:fc] = 0
                else:
                    a_vector[fc:] = 0
            a_curve = ElutionCurve(a_vector)
            similarity = CurveSimilarity(a_curve, similarity.x_curve, orig_a_curve=similarity.a_curve)
            score = similarity.compute_whole_similarity()
            print('score=', score)
            self.logger.info("got a revised score %.2g" % score)

        self.similarity_score = score
        return similarity

    def get_flow_changes(self):
        return self.flowchange.get_flow_changes()

    def get_real_flow_changes(self):
        return self.flowchange.get_real_flow_changes()

    def get_default_angle_range(self, sd):
        from molass_legacy.DataStructure.LPM import get_corrected_matrix        # moved due to ImportError: cannot import name 'get_corrected_matrix' from partially initialized module 'LPM' (most likely due to a circular import)

        D, E, qv, e_curve = sd.get_xr_data_separate_ly()

        # angle_start, flange_limit, pre_rg = get_default_angle_range_impl(D, E, e_curve, qv, sd.xray_index, self.logger)
        qlimit = sd.xr_index

        flange_limit = get_flange_limit(D, E, e_curve, qv)

        change_baseline_correction_defaults(sd, self.flowchange, self.logger)
        correctedD = get_corrected_matrix(D)

        pre_rg = PreliminaryRg(correctedD, E, e_curve, qv, flange_limit)

        angle_start = get_small_anlge_limit(pre_rg, correctedD, E, e_curve, qv, qlimit, self.logger)

        self.angle_start = angle_start
        self.flange_limit = flange_limit
        self.pre_rg = pre_rg
        self.Rg = pre_rg.Rg
        return angle_start, flange_limit

    def get_angle_range(self):
        return self.angle_start, self.flange_limit

    def get_gunier_interval(self):
        sg = self.pre_rg.sg
        return sg.guinier_start, sg.guinier_stop

    def get_rg(self):
        """
        simpler solution is needed,
        because of the trouble with 20180206_TG
        """
        ret_rg = self.Rg
        if ret_rg is None:
            ret_rg = self.pre_rg.compute_rg()
            self.Rg = ret_rg
        return ret_rg

    def set_restrict_info(self, sd, debug=False, figfile=None):
        xr_restrict_list = get_setting('xr_restrict_list')
        uv_restrict_list = get_setting('uv_restrict_list')

        if xr_restrict_list is not None or uv_restrict_list is not None:
            return

        xray_elution_size = sd.intensity_array.shape[0]
        angle_size = sd.intensity_array.shape[1]
        uv_elution_size = sd.conc_array.shape[1]

        angle_start, angle_end = self.get_default_angle_range(sd)

        if angle_start == 0 and angle_end is None:
            xr_angle_restrict = None
        else:
            xr_angle_restrict = TrimmingInfo(1, angle_start, angle_end, angle_size)

        if sd.mtd_elution is None:
            fc = self.flowchange.get_real_flow_changes()
        else:
            fc = sd.mtd_elution.get_linear_slope_ends()

        if fc[0] is None and fc[1] is None:
            uv_restrict_list = None
            xr_elution_restrict = None
        else:
            if sd.mtd_elution is None:
                narrow_trimming = get_setting('narrow_trimming')
                if narrow_trimming:
                    jj = self.flowchange.get_safe_flow_changes()
                    uv_restrict_list = [TrimmingInfo(1, jj[0], jj[1]+1, uv_elution_size), None]
                else:
                    jj = self.flowchange.get_flow_changes()
            else:
                jj = sd.mtd_elution.get_linear_slope_ends()

            # print('self.similarity_score=', self.similarity_score)
            # if self.similarity_score >= ACCEPTABLE_SIMILARITY_SCORE or True:
            if True:
                # TODO: better control for 20180304
                ii = self.get_restrict_list_from_uv_flow_changes(jj, len(sd.absorbance.a_vector), len(sd.xray_curve.x))
                xr_elution_restrict = TrimmingInfo(1, ii[0], ii[1]+1, xray_elution_size)
            else:
                # 20170304
                xr_elution_restrict = None

        xr_restrict_list = [xr_elution_restrict, xr_angle_restrict]

        use_moment_trimming = get_setting('use_moment_trimming')
        if use_moment_trimming:
            from molass_legacy.Trimming.MomentTrimming import get_moment_trimming_info
            from molass_legacy.Trimming.TrimmingUtils import get_default_uv_wl_trimming
            from molass_legacy.Trimming.AutoRestrictorProxy import AutoRestrictorProxy
            ret_info = get_moment_trimming_info(sd)
            uv_elution_restrict = ret_info[0]
            xr_elution_restrict = ret_info[1]
            self.logger.info("moment_trimming_info has been set as %s in PreliminaryRecognition", str(ret_info))
            if uv_restrict_list is None:
                uv_wl_restrict = get_default_uv_wl_trimming(sd)
                uv_restrict_list = [uv_elution_restrict, uv_wl_restrict]
            else:
                uv_restrict_list[0] = uv_elution_restrict
                if uv_restrict_list[1] is None:
                    uv_wl_restrict = get_default_uv_wl_trimming(sd)
                    uv_restrict_list[1] = uv_wl_restrict

            xr_restrict_list[0] = xr_elution_restrict           
            ar = AutoRestrictorProxy(sd, uv_restrict_list, xr_restrict_list)
        else:
            if sd.mtd_elution is None:
                ar = AutoRestrictor(sd, self.cs)
                uv_restrict_list, xr_restrict_list = ar.get_better_restriction(uv_restrict_list, xr_restrict_list, debug=debug, figfile=figfile)
                if debug:
                    print("********************* uv_restrict_list=", uv_restrict_list)
                    print("********************* xr_restrict_list=", xr_restrict_list)
            else:
                ar = None
        self.ar = ar

        set_setting('uv_restrict_list', uv_restrict_list)
        set_setting('xr_restrict_list', xr_restrict_list)
        self.logger.info("uv_restrict_list has been set to %s in PreliminaryRecognition", str(uv_restrict_list))
        self.logger.info("xr_restrict_list has been set to %s in PreliminaryRecognition", str(xr_restrict_list))
        set_setting('uv_restrict_copy', copy.deepcopy(uv_restrict_list))
        set_setting('xr_restrict_copy', copy.deepcopy(xr_restrict_list))

    def reset_restrict_info(self):
        set_setting('xr_restrict_list', None)
        set_setting('uv_restrict_list', None)
        self.set_restrict_info(self.pre_recog_copy)

    def get_restrict_list_from_uv_flow_changes(self, fc, a_size, x_size):
        ret_list = []
        j_end = a_size - 1
        for j in fc:
            if j == 0:
                i = 0
            elif j == j_end:
                i = x_size - 1
            else:
                i = self.cs.inverse_int_value(j)
            ret_list.append(i)

        if False:
            def mapped_plot_closure(ax):
                y = self.cs.x_curve.y
                for i in ret_list:
                    ax.plot(i, y[i], 'o', color='cyan')
            self.cs.plot_the_mapped_state(mapped_plot_closure)

        return ret_list

    def get_xr_slices(self):
        return get_slices_impl("xr_restrict_list")

    def get_uv_slices(self):
        return get_slices_impl("uv_restrict_list")

    def restrict_info_changed(self):
        uv_restrict_list = get_setting('uv_restrict_list')
        uv_restrict_copy = get_setting('uv_restrict_copy')
        xr_restrict_list = get_setting('xr_restrict_list')
        xr_restrict_copy = get_setting('xr_restrict_copy')
        if (    str(uv_restrict_list) == str(uv_restrict_copy)
            and str(xr_restrict_list) == str(xr_restrict_copy)
            ):
            return False
        else:
            return True

    def get_pre_recog_copy(self):
        print('pre_recog_copy.absorbance=', self.pre_recog_copy.absorbance)
        return  self.pre_recog_copy

    def get_xray_shift(self):
        restrict = get_setting('xr_restrict_list')[0]
        shift = 0 if restrict is None else restrict.start
        if shift is None:
            shift = 0
        return shift

    def get_restricted_mapped_info(self):
        best_params, max_score_pair, max_info = self.mapped_info
        A, B = best_params
        shift = self.get_xray_shift()
        res_params = np.array( [A, B-shift])

        if False:
            import molass_legacy.KekLib.DebugPlot as plt
            uv_curve = self.pre_recog_copy.absorbance.a_curve
            xray_curve = self.pre_recog_copy.xray_curve
            fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(14,7))
            ax1, ax2 = axes
            ax1.plot(uv_curve.x, uv_curve.y, color='blue')
            ax2.plot(xray_curve.x, xray_curve.y, color='orange')
            if restrict is not None:
                ymin, ymax = ax2.get_ylim()
                ax2.set_ylim(ymin, ymax)
                for p in [restrict.start, restrict.stop]:
                    ax2.plot([p, p], [ymin, ymax], ':', color='gray')
            fig.tight_layout()
            plt.show()

        return [res_params, max_score_pair, max_info]



# coding: utf-8
"""
    TwoStateSolver.py

    Copyright (c) 2019,2024, SAXS Team, KEK-PF
"""
import os
import time
import numpy as np
from scipy import optimize, stats
from lmfit import Parameters, minimize
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.SciPyCookbook import smooth
from SimpleUnfolding import proportion_folded
from SvdDenoise import get_denoised_data

class TwoStateCurves:
    def __init__(self, G, m, ab_params):
        self.G = G
        self.m = m
        self.ab_params = ab_params

    def compute_yf(self, x):
        pf = proportion_folded(self.G, self.m, x)
        af, bf = self.ab_params[0:2]
        return af*pf + bf*x*pf

    def compute_yu(self, x):
        pf = proportion_folded(self.G, self.m, x)
        pu = 1 - pf
        au, bu = self.ab_params[2:4]
        return au*pu + bu*x*pu

class TwoStateSolver:
    def __init__(self):
        pass

    def solve(self, data):
        print(data.shape)
        M = get_denoised_data(data, rank=4)
        x = np.arange(M.shape[1])

        def compute_norm(params):
            G, m = params
            # print('G, m=', G, m)
            pf = proportion_folded(G, m, x)
            pu = 1 - pf
            C = np.array([pf, x*pf, pu, x*pu])
            Cpinv = np.linalg.pinv(C)
            P = M@Cpinv
            z = np.linalg.norm(P@C - M)
            return z

        def get_init_params(G):
            G_init = G
            mx = (x[0] + x[-1])/2
            m_init = G_init/mx
            return G_init, m_init

        if True:
            t0 = time.time()
            min_norm = None
            opt_x = None
            # for g in [0, 0.1, 0.2] + list(range(1, 11)):
            # for g in [0, 0.1, 0.2, 1]:
            for g in range(1, 11):
                G_init, m_init = get_init_params(g)
                res= optimize.minimize(compute_norm, [G_init, m_init], bounds=((0,10),(0,1)))
                norm = compute_norm(res.x)
                # print([g], norm, res.x)
                if min_norm is None or norm < min_norm:
                    min_norm = norm
                    opt_x = res.x
            self.params = opt_x
            print("It took", time.time() - t0)
        else:
            from SimulatedAnnealing import SimulatedAnnealing
            G_init, m_init = get_init_params(3)
            anneal = SimulatedAnnealing()
            anneal.minimize( compute_norm, xrange=[[0, 10], [0, 1]], start=[G_init, m_init], seed=1234,
                            xconstaints=None )
            self.params = anneal.xc
        self.M = M
        self.compute_components(x)

    def compute_components(self, x):
        G, m = self.params
        print('G, m=', G, m)
        pf = proportion_folded(G, m, x)
        pu = 1 - pf
        C = np.array([pf, x*pf, pu, x*pu])
        Cpinv = np.linalg.pinv(C)
        P = self.M@Cpinv
        self.C = C
        self.P = P

    def plot_components(self, ax, data, index, title=None, axt=None, start=0):
        draw_coeff=False
        x = np.arange(start, start+data.shape[1])
        y = data[index,:]
        self.x = x
        self.y = y
        C = self.C
        pf  = C[0,:]
        xpf = C[1,:]
        pu  = C[2,:]
        xpu = C[3,:]
        af, bf, au, bu = self.P[index,:]
        yf = af*pf + bf*xpf
        yu = au*pu + bu*xpu

        if axt is None:
            axt = ax.twinx()
        axt.grid(False)

        if title is not None:
            ax.set_title(title, fontsize=16)
        ax.set_xlabel('Eno')
        ax.set_ylabel('Proportion')
        axt.set_ylabel('Intensity')

        G, m = self.params
        ax.plot(x, pf, label='$P_{F;G=%2.g,m=%.2g}$' % (G, m))
        ax.plot(x, pu, label='$P_U$')
        axt.plot(x, y, color='orange', label='given')
        axt.plot(x, yf, ':', label='$y_F$')
        axt.plot(x, yu, ':', label='$y_U$')
        axt.plot(x, yf+yu, ':', color='orange', label='$y_F+y_U$')

        bbox = (0, 0.65) if draw_coeff else None
        ax.legend(bbox_to_anchor=bbox, loc='center left', fontsize=16)
        axt.legend(loc='upper right', fontsize=16)

    def draw_components_with_bands(self, ax, data, index, title=None, start=0, detail=False):
        from matplotlib.patches import Polygon
        from molass_legacy.KekLib.OurMatplotlib import get_color
        from GeometryUtils import polygon_area_centroid

        x = np.arange(start, start+data.shape[1])
        y = data[index,:]
        self.x = x
        self.y = y
        C = self.C
        pf  = C[0,:]
        xpf = C[1,:]
        pu  = C[2,:]
        xpu = C[3,:]
        af, bf, au, bu = self.P[index,:]
        yf = af*pf + bf*xpf
        yu = au*pu + bu*xpu
        ax.set_xlabel('Eno')
        ax.set_ylabel('Proportion')
        ax.plot(x, y, color='orange', label='given')
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(0, ymax)

        poly_points = list(zip( x, yf )) + [(x[-1], 0), (x[0], 0)]
        f_color = get_color(0)
        f_band = Polygon( poly_points, color=f_color, alpha=0.2 )
        ax.add_patch(f_band)
        cp = polygon_area_centroid(f_band)
        ax.text(*cp, "F", color=f_color, ha='center', va='center', fontsize=50)

        if detail:
            poly_points = list(zip( x, yf )) + list( reversed( list( zip( x, yf - bf*xpf ) ) ) )
            color='pink'
            d_band = Polygon( poly_points, color=color, alpha=0.2 )
            ax.add_patch(d_band)
            cp = polygon_area_centroid(d_band)
            ax.text(*cp, "F.d.d.", color=color, ha='center', va='center', fontsize=50)

        poly_points = list(zip( x, yf )) + list( reversed( list( zip( x, yf+yu ) ) ) )
        u_color = get_color(1)
        u_band = Polygon( poly_points, color=u_color, alpha=0.2 )
        ax.add_patch(u_band)
        cp = polygon_area_centroid(u_band)
        ax.text(*cp, "U", color=u_color, ha='center', va='center', fontsize=50)

    def guess_blank_location(self, slice_):
        G, m = self.params
        cx = G/m
        xsize = slice_.stop - slice_.start
        pos_ratio = cx/xsize
        slope, intercept, r_value, p_value, std_err = stats.linregress(self.x, self.y)

        # print('pos_ratio=', pos_ratio, 'slope=', slope)

        if pos_ratio < 0.4  or (pos_ratio < 0.6 and slope > 0):
            side = 'right'
        else:
            side = 'left'
        return side

    def get_crossing_point(self):
        G, m = self.params
        cp = G/m
        # print('G=', G, 'm=', m, 'cp=', cp)
        return int(cp)

    def get_modeled_curves(self, index):
        curves = TwoStateCurves(*self.params, self.P[index,:])
        return curves.compute_yf, curves.compute_yu



"""
    ModelParams.SdmParamsSheet.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.KekLib.OurTkinter import Tk
from tksheet import Sheet
from importlib import reload
import molass_legacy.ModelParams.ParamsSheetBase
reload(molass_legacy.ModelParams.ParamsSheetBase)
from .ParamsSheetBase import ParamsSheetBase

class SdmParamsSheet(ParamsSheetBase):
    def __init__(self, parent, params, dsets, optimizer):
        ParamsSheetBase.__init__(self, parent, params, dsets, optimizer)

        self.n = optimizer.n_components
        nc  = self.n - 1
        self.params_addr = [None]*len(params)
        self.params_addr_inv = {}

        body_frame = Tk.Frame(self)
        body_frame.pack()

        xr_params, xr_baseparams, rgs, mapping, uv_params, uv_baseparams, mappable_range, sdmcol_params = optimizer.split_params_simple(params)
        N, K, t0, rp, N0, tI = sdmcol_params
        T = K/N
        me = 1.5
        mp = 1.5
    
        xr_base = 0
        xr_bp_base = xr_base + np.product(xr_params.shape)
        rg_base = xr_bp_base + len(xr_baseparams)
        mp_base = rg_base + len(rgs)
        uv_base = mp_base + len(mapping)
        uv_bp_base = uv_base + len(uv_params)
        mr_base = uv_bp_base + len(uv_baseparams)
        se_base = mr_base + len(mappable_range)

        if dsets is None:
            num_columns = 8
            extra_headers = []
            xr_proportions = None
            uv_proportions = None
        else:
            num_columns = 10
            extra_headers = ["", "", "proportion"]

            opt_lrf_info = optimizer.objective_func(params, return_lrf_info=True)
            xr_proportions = opt_lrf_info.get_xr_proportions()
            uv_proportions = opt_lrf_info.get_uv_proportions()

        K = N*T
        m = me + mp
        sec_params_disp = (rp, t0, tI, N0, K, m)
        # in C1100.py
        # N, K, x0, poresize, N0, tI = sdmcol_params
        sec_params_disp_map = [3,   # rp
                               2,   # t0
                               5,   # tI
                               4,   # N0
                               1,   # K
                               None,# m
                               ]
        num_extended_rows = len(sec_params_disp) + 1

        data_list = [["" for c in range(num_columns)] for r in range(self.n*2 + 2 + 8 + num_extended_rows)]

        row_offset = 0
        for j, name in enumerate(["h", "", "", "", "", "rg"] + extra_headers, start=1):
            data_list[row_offset][j] = name

        row_offset += 1
        n = xr_params.shape[0]
        xr_col_size = 4     # xr_params.shape[1] for egh
        for i in range(n):
            if i == 0:
                v = "xr_params"
            else:
                v = ""
            data_list[row_offset+i][0] = v
            data_list[row_offset+i][1] = "%g" % xr_params[i]
            self.set_params_addr(xr_base+i, (row_offset+i, 1))

            data_list[row_offset+i][xr_col_size+2] = "%g" % rgs[i]
            self.set_params_addr(rg_base+i, (row_offset+i, xr_col_size+2))

            if xr_proportions is not None:
                if i == 0:
                    data_list[row_offset+i][xr_col_size+4] = "xr area"
                if i < len(xr_proportions):
                    data_list[row_offset+i][xr_col_size+5] = "%g" % xr_proportions[i]

        row_offset += n + 1
        col_names = ["slope", "intercept"]
        if len(xr_baseparams) == 3:
            col_names.append("fouling")

        for j, name in enumerate(col_names):
            data_list[row_offset][1+j] = name

        row_offset += 1
        for i in range(1):
            if i == 0:
                v = "xr_baseline"
            else:
                v = ""
            data_list[row_offset+i][0] = v
            num_params = len(xr_baseparams)
            for j in range(num_params):
                data_list[row_offset+i][j+1] = "%g" % xr_baseparams[j]
                self.set_params_addr(xr_bp_base+j, (row_offset+i, j+1))

            # data_list[row_offset+i][xr_col_size+2] = "%g" % rgs[n+i]
            # self.set_params_addr(rg_base+n+i, (row_offset+i, xr_col_size+2))

        row_offset += 2
        data_list[row_offset][1] = "h"
        if uv_proportions is not None:
            data_list[row_offset][9] = "proportion"

        row_offset += 1
        for i in range(n):
            if i == 0:
                v = "uv_params"
            else:
                v = ""
            data_list[row_offset+i][0] = v
            data_list[row_offset+i][1] = "%g" % uv_params[i]
            self.set_params_addr(uv_base+i, (row_offset+i, 1))

            if uv_proportions is not None:
                if i == 0:
                    data_list[row_offset+i][xr_col_size+4] = "uv area"
                if i < len(uv_proportions):
                    data_list[row_offset+i][xr_col_size+5] = "%g" % uv_proportions[i]

        row_offset += n + 1
        col_names = ["L", "x0", "k", "b", "s1", "s2", "diff_ratio"]
        if len(uv_baseparams) == 8:
            col_names.append("fouling")

        for j, name in enumerate(col_names):
            data_list[row_offset][1+j] = name

        row_offset += 1
        for i in range(1):
            if i == 0:
                v = "uv_baseline"
            else:
                v = ""
            data_list[row_offset+i][0] = v
            num_params = len(uv_baseparams)
            for j in range(num_params):
                data_list[row_offset+i][j+1] = "%g" % uv_baseparams[j]
                self.set_params_addr(uv_bp_base+j, (row_offset+i, j+1))

        row_offset += 2
        for j, name in enumerate(["slope", "intercept", "", "", "from", "to"]):
            data_list[row_offset][1+j] = name

        row_offset += 1
        data_list[row_offset][0] = "mapping"
        data_list[row_offset][4] = "mappable_range"
        for j in range(2):
            data_list[row_offset][1+j] = "%g" % mapping[j]
            self.set_params_addr(mp_base+j, (row_offset, 1+j))
            data_list[row_offset][5+j] = "%g" % mappable_range[j]
            self.set_params_addr(mr_base+j, (row_offset, 5+j))

        dark_cells = []

        if sdmcol_params is not None:
            row_offset += 1
            for name, value, i in zip(["rp (pore size)", "t0", "tI", "N0", "K", "m"], sec_params_disp, [0, 1, 2, 3, -1, -1]):
                row_offset += 1
                data_list[row_offset][0] = name
                data_list[row_offset][1] = "%g" % value
                if i >= 0:
                    j = sec_params_disp_map[i]
                    self.set_params_addr(se_base+j, (row_offset, 1))
                else:
                    dark_cells.append((row_offset, 0))
                    dark_cells.append((row_offset, 1))
                    if name == "K":
                        data_list[row_offset][3] = "N (nperm)"
                        data_list[row_offset][4] = "%g" % N
                        self.set_params_addr(se_base, (row_offset, 4))
                        data_list[row_offset][6] = "T (tperm)"
                        data_list[row_offset][7] = "%g" % T
                        self.set_params_addr(se_base+1, (row_offset, 7))    # se_base+1 is actually K
                    elif name == "m":
                        data_list[row_offset][3] = "me"
                        data_list[row_offset][4] = "%g" % me
                        # self.set_params_addr(se_base+3, (row_offset, 4))  # currently no correspondence
                        data_list[row_offset][6] = "mp"
                        data_list[row_offset][7] = "%g" % mp
                        # self.set_params_addr(se_base+5, (row_offset, 7))  # currently no correspondence
                    else:
                        assert False

        self.num_valid_rows = row_offset + 2
        self.data_list = data_list
        column_width = 90
        width = column_width*num_columns + 60
        height = int(25*self.num_valid_rows) + 60
        self.sheet = Sheet(body_frame, width=width, height=height, data=data_list, show_selected_cells_border=False, column_width=column_width)

        for r, c in dark_cells:
            self.sheet.highlight_cells(r, c, bg="gray", fg="white")

        self.sheet.pack()

    def highlight_active_params(self, state_info):
        indeces = self.get_active_indeces(state_info)
        print("indeces=", indeces)
        active_cells = [self.get_cell_address(k) for k in indeces]
        for i, addr in enumerate(active_cells):
            if addr is None:
                continue
            else:
                r, c = addr
            self.sheet.highlight_cells(r, c, bg="pink")


"""
    SimplestMapper.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""

class SimplestMapper:
    def __init__(self, curve1, curve2):
        self.scale = len(curve2.x)/len(curve1.x)
        self.origin1 = curve1.primary_peak_x
        self.origin2 = curve2.primary_peak_x

    def map_1to2(self, i):
        return int(round(self.origin2 + self.scale * (i - self.origin1)))



"""
    list_bestn.py
"""

rmsd_list = []

with open("rmsd.csv") as fh:
    for k, line in enumerate(fh):
        # print(line)
        code, rmsd = line[:-1].split(",")
        v = float(rmsd)
        # print([k], code, v)
        rmsd_list.append((code, v))
        if k > 3:
            # break
            pass

rmsd_list = sorted(rmsd_list, key=lambda x: x[1])

for pair in rmsd_list[0:5]:
    print(*pair)
